# Copyright 2020 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
""" Quantized Recurrent layers. """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import warnings
import tensorflow as tf
from tensorflow.keras import activations
from tensorflow.keras import constraints
from tensorflow.keras import initializers
from tensorflow.keras import regularizers
from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import SimpleRNNCell
from tensorflow.keras.layers import LSTMCell
from tensorflow.keras.layers import GRUCell
from tensorflow.keras.layers import RNN
from tensorflow.keras.layers import Bidirectional
from tensorflow.python.util import nest
from tensorflow.python.ops import array_ops
from tensorflow.python.framework import ops
from tensorflow_model_optimization.python.core.sparsity.keras.prunable_layer import PrunableLayer

import tensorflow.keras.backend as K
from .qlayers import get_auto_range_constraint_initializer
from .qlayers import QActivation
from .quantizers import get_quantized_initializer
from .quantizers import get_quantizer


class QSimpleRNNCell(SimpleRNNCell):
  """
  Cell class for the QSimpleRNNCell layer.

  Most of these parameters follow the implementation of SimpleRNNCell in
  Keras, with the exception of kernel_quantizer, recurrent_quantizer,
  bias_quantizer, and state_quantizer.

  kernel_quantizer: quantizer function/class for kernel
  recurrent_quantizer: quantizer function/class for recurrent kernel
  bias_quantizer: quantizer function/class for bias
  state_quantizer: quantizer function/class for states

  We refer the reader to the documentation of SimpleRNNCell in Keras for the
  other parameters.

  """
  def __init__(self,
               units,
               activation='quantized_tanh',
               use_bias=True,
               kernel_initializer='glorot_uniform',
               recurrent_initializer='orthogonal',
               bias_initializer='zeros',
               kernel_regularizer=None,
               recurrent_regularizer=None,
               bias_regularizer=None,
               kernel_constraint=None,
               recurrent_constraint=None,
               bias_constraint=None,
               kernel_quantizer=None,
               recurrent_quantizer=None,
               bias_quantizer=None,
               state_quantizer=None,
               dropout=0.,
               recurrent_dropout=0.,
               **kwargs):

    self.kernel_quantizer = kernel_quantizer
    self.recurrent_quantizer = recurrent_quantizer
    self.bias_quantizer = bias_quantizer
    self.state_quantizer = state_quantizer

    self.kernel_quantizer_internal = get_quantizer(self.kernel_quantizer)
    self.recurrent_quantizer_internal = get_quantizer(self.recurrent_quantizer)
    self.bias_quantizer_internal = get_quantizer(self.bias_quantizer)
    self.state_quantizer_internal = get_quantizer(self.state_quantizer)

    self.quantizers = [
        self.kernel_quantizer_internal, self.recurrent_quantizer_internal,
        self.bias_quantizer_internal, self.state_quantizer_internal
    ]

    if hasattr(self.kernel_quantizer_internal, "_set_trainable_parameter"):
      self.kernel_quantizer_internal._set_trainable_parameter()

    if hasattr(self.recurrent_quantizer_internal, "_set_trainable_parameter"):
      self.recurrent_quantizer_internal._set_trainable_parameter()

    kernel_constraint, kernel_initializer = (
        get_auto_range_constraint_initializer(self.kernel_quantizer_internal,
                                              kernel_constraint,
                                              kernel_initializer))

    recurrent_constraint, recurrent_initializer = (
        get_auto_range_constraint_initializer(self.recurrent_quantizer_internal,
                                              recurrent_constraint,
                                              recurrent_initializer))

    if use_bias:
      bias_constraint, bias_initializer = (
          get_auto_range_constraint_initializer(self.bias_quantizer_internal,
                                                bias_constraint,
                                                bias_initializer))

    if activation is not None:
      activation = get_quantizer(activation)

    super(QSimpleRNNCell, self).__init__(
      units=units,
      activation=activation,
      use_bias=use_bias,
      kernel_initializer=kernel_initializer,
      recurrent_initializer=recurrent_initializer,
      bias_initializer=bias_initializer,
      kernel_regularizer=kernel_regularizer,
      recurrent_regularizer=recurrent_regularizer,
      bias_regularizer=bias_regularizer,
      kernel_constraint=kernel_constraint,
      recurrent_constraint=recurrent_constraint,
      bias_constraint=bias_constraint,
      dropout=dropout,
      recurrent_dropout=recurrent_dropout,
      **kwargs
    )

  def call(self, inputs, states, training=None):
    prev_output = states[0] if nest.is_sequence(states) else states

    dp_mask = self.get_dropout_mask_for_cell(inputs, training)
    rec_dp_mask = self.get_recurrent_dropout_mask_for_cell(
        prev_output, training)

    if self.state_quantizer:
      quantized_prev_output = self.state_quantizer_internal(prev_output)
    else:
      quantized_prev_output = prev_output

    if self.kernel_quantizer:
      quantized_kernel = self.kernel_quantizer_internal(self.kernel)
    else:
      quantized_kernel = self.kernel

    if dp_mask is not None:
      h = K.dot(inputs * dp_mask, quantized_kernel)
    else:
      h = K.dot(inputs, quantized_kernel)

    if self.bias is not None:
      if self.bias_quantizer:
        quantized_bias = self.bias_quantizer_internal(self.bias)
      else:
        quantized_bias = self.bias

      h = K.bias_add(h, quantized_bias)

    if rec_dp_mask is not None:
      quantized_prev_output = quantized_prev_output * rec_dp_mask

    if self.recurrent_quantizer:
      quantized_recurrent = self.recurrent_quantizer_internal(self.recurrent_kernel)
    else:
      quantized_recurrent = self.recurrent_kernel

    output = h + K.dot(quantized_prev_output, quantized_recurrent)

    if self.activation is not None:
      output = self.activation(output)
    return output, [output]

  def get_config(self):
    config = {
        "kernel_quantizer":
            constraints.serialize(self.kernel_quantizer_internal),
        "recurrent_quantizer":
            constraints.serialize(self.recurrent_quantizer_internal),
        "bias_quantizer":
            constraints.serialize(self.bias_quantizer_internal),
        "state_quantizer":
            constraints.serialize(self.state_quantizer_internal)
    }
    base_config = super(QSimpleRNNCell, self).get_config()
    return dict(list(base_config.items()) + list(config.items()))


class QSimpleRNN(RNN, PrunableLayer):
  """
  Class for the QSimpleRNN layer.

  Most of these parameters follow the implementation of SimpleRNN in
  Keras, with the exception of kernel_quantizer, recurrent_quantizer,
  bias_quantizer and state_quantizer.


  kernel_quantizer: quantizer function/class for kernel
  recurrent_quantizer: quantizer function/class for recurrent kernel
  bias_quantizer: quantizer function/class for bias
  state_quantizer: quantizer function/class for states


  We refer the reader to the documentation of SimpleRNN in Keras for the
  other parameters.

  """

  def __init__(self,
               units,
               activation='quantized_tanh',
               use_bias=True,
               kernel_initializer='glorot_uniform',
               recurrent_initializer='orthogonal',
               bias_initializer='zeros',
               kernel_regularizer=None,
               recurrent_regularizer=None,
               bias_regularizer=None,
               activity_regularizer=None,
               kernel_constraint=None,
               recurrent_constraint=None,
               bias_constraint=None,
               kernel_quantizer=None,
               recurrent_quantizer=None,
               bias_quantizer=None,
               state_quantizer=None,
               dropout=0.,
               recurrent_dropout=0.,
               return_sequences=False,
               return_state=False,
               go_backwards=False,
               stateful=False,
               unroll=False,
               **kwargs):

    if 'enable_caching_device' in kwargs:
      cell_kwargs = {'enable_caching_device':
                     kwargs.pop('enable_caching_device')}
    else:
      cell_kwargs = {}

    cell = QSimpleRNNCell(
        units,
        activation=activation,
        use_bias=use_bias,
        kernel_initializer=kernel_initializer,
        recurrent_initializer=recurrent_initializer,
        bias_initializer=bias_initializer,
        kernel_regularizer=kernel_regularizer,
        recurrent_regularizer=recurrent_regularizer,
        bias_regularizer=bias_regularizer,
        kernel_constraint=kernel_constraint,
        recurrent_constraint=recurrent_constraint,
        bias_constraint=bias_constraint,
        kernel_quantizer=kernel_quantizer,
        recurrent_quantizer=recurrent_quantizer,
        bias_quantizer=bias_quantizer,
        state_quantizer=state_quantizer,
        dropout=dropout,
        recurrent_dropout=recurrent_dropout,
        dtype=kwargs.get('dtype'),
        trainable=kwargs.get('trainable', True),
        **cell_kwargs)

    super(QSimpleRNN, self).__init__(
        cell,
        return_sequences=return_sequences,
        return_state=return_state,
        go_backwards=go_backwards,
        stateful=stateful,
        unroll=unroll,
        **kwargs)
    self.activity_regularizer = regularizers.get(activity_regularizer)
    self.input_spec = [tf.keras.layers.InputSpec(ndim=3)]

  def call(self, inputs, mask=None, training=None, initial_state=None):
    self._maybe_reset_cell_dropout_mask(self.cell)
    return super(QSimpleRNN, self).call(
        inputs, mask=mask, training=training, initial_state=initial_state)

  def get_quantizers(self):
    return self.cell.quantizers

  def get_prunable_weights(self):
    return [self.cell.kernel, self.cell.recurrent_kernel]

  @property
  def units(self):
    return self.cell.units

  @property
  def activation(self):
    return self.cell.activation

  @property
  def use_bias(self):
    return self.cell.use_bias

  @property
  def kernel_initializer(self):
    return self.cell.kernel_initializer

  @property
  def recurrent_initializer(self):
    return self.cell.recurrent_initializer

  @property
  def bias_initializer(self):
    return self.cell.bias_initializer

  @property
  def kernel_regularizer(self):
    return self.cell.kernel_regularizer

  @property
  def recurrent_regularizer(self):
    return self.cell.recurrent_regularizer

  @property
  def bias_regularizer(self):
    return self.cell.bias_regularizer

  @property
  def kernel_constraint(self):
    return self.cell.kernel_constraint

  @property
  def recurrent_constraint(self):
    return self.cell.recurrent_constraint

  @property
  def bias_constraint(self):
    return self.cell.bias_constraint

  @property
  def kernel_quantizer_internal(self):
    return self.cell.kernel_quantizer_internal

  @property
  def recurrent_quantizer_internal(self):
    return self.cell.recurrent_quantizer_internal

  @property
  def bias_quantizer_internal(self):
    return self.cell.bias_quantizer_internal

  @property
  def state_quantizer_internal(self):
    return self.cell.state_quantizer_internal

  @property
  def kernel_quantizer(self):
    return self.cell.kernel_quantizer

  @property
  def recurrent_quantizer(self):
    return self.cell.recurrent_quantizer

  @property
  def bias_quantizer(self):
    return self.cell.bias_quantizer

  @property
  def state_quantizer(self):
    return self.cell.state_quantizer

  @property
  def dropout(self):
    return self.cell.dropout

  @property
  def recurrent_dropout(self):
    return self.cell.recurrent_dropout

  def get_config(self):
    config = {
        'units':
            self.units,
        'activation':
            activations.serialize(self.activation),
        'use_bias':
            self.use_bias,
        'kernel_initializer':
            initializers.serialize(self.kernel_initializer),
        'recurrent_initializer':
            initializers.serialize(self.recurrent_initializer),
        'bias_initializer':
            initializers.serialize(self.bias_initializer),
        'kernel_regularizer':
            regularizers.serialize(self.kernel_regularizer),
        'recurrent_regularizer':
            regularizers.serialize(self.recurrent_regularizer),
        'bias_regularizer':
            regularizers.serialize(self.bias_regularizer),
        'activity_regularizer':
            regularizers.serialize(self.activity_regularizer),
        'kernel_constraint':
            constraints.serialize(self.kernel_constraint),
        'recurrent_constraint':
            constraints.serialize(self.recurrent_constraint),
        'bias_constraint':
            constraints.serialize(self.bias_constraint),
        "kernel_quantizer":
            constraints.serialize(self.kernel_quantizer_internal),
        "recurrent_quantizer":
            constraints.serialize(self.recurrent_quantizer_internal),
        "bias_quantizer":
            constraints.serialize(self.bias_quantizer_internal),
        "state_quantizer":
            constraints.serialize(self.state_quantizer_internal),
        'dropout':
            self.dropout,
        'recurrent_dropout':
            self.recurrent_dropout
    }
    base_config = super(QSimpleRNN, self).get_config()
    del base_config['cell']
    return dict(list(base_config.items()) + list(config.items()))

  def get_quantization_config(self):
    return {
        "kernel_quantizer":
            str(self.kernel_quantizer_internal),
        "recurrent_quantizer":
            str(self.recurrent_quantizer_internal),
        "bias_quantizer":
            str(self.bias_quantizer_internal),
        "state_quantizer":
            str(self.state_quantizer_internal),
        "activation":
            str(self.activation)
    }

  @classmethod
  def from_config(cls, config):
    if 'implementation' in config:
      config.pop('implementation')
    return cls(**config)


class QLSTMCell(LSTMCell):
  """
  Cell class for the QLSTMCell layer.

  Most of these parameters follow the implementation of LSTMCell in
  Keras, with the exception of kernel_quantizer, recurrent_quantizer,
  bias_quantizer, state_quantizer.


  kernel_quantizer: quantizer function/class for kernel
  recurrent_quantizer: quantizer function/class for recurrent kernel
  bias_quantizer: quantizer function/class for bias
  state_quantizer: quantizer function/class for states

  We refer the reader to the documentation of LSTMCell in Keras for the
  other parameters.

  """

  def __init__(self,
               units,
               activation='quantized_tanh',
               recurrent_activation='hard_sigmoid',
               use_bias=True,
               kernel_initializer='glorot_uniform',
               recurrent_initializer='orthogonal',
               bias_initializer='zeros',
               unit_forget_bias=True,
               kernel_regularizer=None,
               recurrent_regularizer=None,
               bias_regularizer=None,
               kernel_constraint=None,
               recurrent_constraint=None,
               bias_constraint=None,
               kernel_quantizer=None,
               recurrent_quantizer=None,
               bias_quantizer=None,
               state_quantizer=None,
               dropout=0.,
               recurrent_dropout=0.,
               implementation=1,
               **kwargs):
    self.kernel_quantizer = kernel_quantizer
    self.recurrent_quantizer = recurrent_quantizer
    self.bias_quantizer = bias_quantizer
    self.state_quantizer = state_quantizer

    self.kernel_quantizer_internal = get_quantizer(self.kernel_quantizer)
    self.recurrent_quantizer_internal = get_quantizer(self.recurrent_quantizer)
    self.bias_quantizer_internal = get_quantizer(self.bias_quantizer)
    self.state_quantizer_internal = get_quantizer(self.state_quantizer)

    self.quantizers = [
      self.kernel_quantizer_internal,
      self.recurrent_quantizer_internal,
      self.bias_quantizer_internal,
      self.state_quantizer_internal,
    ]

    if hasattr(self.kernel_quantizer_internal, "_set_trainable_parameter"):
      self.kernel_quantizer_internal._set_trainable_parameter()

    if hasattr(self.recurrent_quantizer_internal, "_set_trainable_parameter"):
      self.recurrent_quantizer_internal._set_trainable_parameter()

    kernel_constraint, kernel_initializer = (
        get_auto_range_constraint_initializer(self.kernel_quantizer_internal,
                                              kernel_constraint,
                                              kernel_initializer))

    recurrent_constraint, recurrent_initializer = (
        get_auto_range_constraint_initializer(self.recurrent_quantizer_internal,
                                              recurrent_constraint,
                                              recurrent_initializer))

    if use_bias:
      bias_constraint, bias_initializer = (
          get_auto_range_constraint_initializer(self.bias_quantizer_internal,
                                                bias_constraint,
                                                bias_initializer))

    if activation is not None:
      activation = get_quantizer(activation)

    if recurrent_activation is not None:
      recurrent_activation = get_quantizer(recurrent_activation)

    super(QLSTMCell, self).__init__(
      units=units,
      activation=activation,
      use_bias=use_bias,
      recurrent_activation=recurrent_activation,
      kernel_initializer=kernel_initializer,
      recurrent_initializer=recurrent_initializer,
      bias_initializer=bias_initializer,
      unit_forget_bias=True,
      kernel_regularizer=kernel_regularizer,
      recurrent_regularizer=recurrent_regularizer,
      bias_regularizer=bias_regularizer,
      kernel_constraint=kernel_constraint,
      recurrent_constraint=recurrent_constraint,
      bias_constraint=bias_constraint,
      dropout=dropout,
      recurrent_dropout=recurrent_dropout,
      implementation=implementation,
      **kwargs
    )

  def _compute_carry_and_output(self, x, h_tm1, c_tm1, quantized_recurrent):
    """Computes carry and output using split kernels."""
    x_i, x_f, x_c, x_o = x
    h_tm1_i, h_tm1_f, h_tm1_c, h_tm1_o = h_tm1
    i = self.recurrent_activation(
        x_i + K.dot(h_tm1_i, quantized_recurrent[:, :self.units]))
    f = self.recurrent_activation(x_f + K.dot(
        h_tm1_f, quantized_recurrent[:, self.units:self.units * 2]))
    c = f * c_tm1 + i * self.activation(x_c + K.dot(
        h_tm1_c, quantized_recurrent[:, self.units * 2:self.units * 3]))
    o = self.recurrent_activation(
        x_o + K.dot(h_tm1_o, quantized_recurrent[:, self.units * 3:]))
    return c, o

  def _compute_carry_and_output_fused(self, z, c_tm1):
    """Computes carry and output using fused kernels."""
    z0, z1, z2, z3 = z
    i = self.recurrent_activation(z0)
    f = self.recurrent_activation(z1)
    c = f * c_tm1 + i * self.activation(z2)
    o = self.recurrent_activation(z3)
    return c, o

  def call(self, inputs, states, training=None):
    h_tm1_tmp = states[0]  # previous memory state
    c_tm1_tmp = states[1]  # previous carry state

    if self.state_quantizer:
      c_tm1 = self.state_quantizer_internal(c_tm1_tmp)
      h_tm1 = self.state_quantizer_internal(h_tm1_tmp)
    else:
      c_tm1 = c_tm1_tmp
      h_tm1 = h_tm1_tmp

    dp_mask = self.get_dropout_mask_for_cell(inputs, training, count=4)
    rec_dp_mask = self.get_recurrent_dropout_mask_for_cell(
        h_tm1, training, count=4)

    if self.kernel_quantizer:
      quantized_kernel = self.kernel_quantizer_internal(self.kernel)
    else:
      quantized_kernel = self.kernel
    if self.recurrent_quantizer:
      quantized_recurrent = self.recurrent_quantizer_internal(self.recurrent_kernel)
    else:
      quantized_recurrent = self.recurrent_kernel
    if self.bias_quantizer:
      quantized_bias = self.bias_quantizer_internal(self.bias)
    else:
      quantized_bias = self.bias

    if self.implementation == 1:
      if 0 < self.dropout < 1.:
        inputs_i = inputs * dp_mask[0]
        inputs_f = inputs * dp_mask[1]
        inputs_c = inputs * dp_mask[2]
        inputs_o = inputs * dp_mask[3]
      else:
        inputs_i = inputs
        inputs_f = inputs
        inputs_c = inputs
        inputs_o = inputs
      k_i, k_f, k_c, k_o = array_ops.split(
          quantized_kernel, num_or_size_splits=4, axis=1)
      x_i = K.dot(inputs_i, k_i)
      x_f = K.dot(inputs_f, k_f)
      x_c = K.dot(inputs_c, k_c)
      x_o = K.dot(inputs_o, k_o)
      if self.use_bias:
        b_i, b_f, b_c, b_o = array_ops.split(
            quantized_bias, num_or_size_splits=4, axis=0)
        x_i = K.bias_add(x_i, b_i)
        x_f = K.bias_add(x_f, b_f)
        x_c = K.bias_add(x_c, b_c)
        x_o = K.bias_add(x_o, b_o)

      if 0 < self.recurrent_dropout < 1.:
        h_tm1_i = h_tm1 * rec_dp_mask[0]
        h_tm1_f = h_tm1 * rec_dp_mask[1]
        h_tm1_c = h_tm1 * rec_dp_mask[2]
        h_tm1_o = h_tm1 * rec_dp_mask[3]
      else:
        h_tm1_i = h_tm1
        h_tm1_f = h_tm1
        h_tm1_c = h_tm1
        h_tm1_o = h_tm1
      x = (x_i, x_f, x_c, x_o)
      h_tm1 = (h_tm1_i, h_tm1_f, h_tm1_c, h_tm1_o)
      c, o = self._compute_carry_and_output(x, h_tm1, c_tm1, quantized_recurrent)
    else:
      if 0. < self.dropout < 1.:
        inputs = inputs * dp_mask[0]
      z = K.dot(inputs, quantized_kernel)
      z += K.dot(h_tm1, quantized_recurrent)
      if self.use_bias:
        z = K.bias_add(z, quantized_bias)

      z = array_ops.split(z, num_or_size_splits=4, axis=1)
      c, o = self._compute_carry_and_output_fused(z, c_tm1)

    h = o * self.activation(c)
    return h, [h, c]

  def get_config(self):
    config = {
        "kernel_quantizer":
            constraints.serialize(self.kernel_quantizer_internal),
        "recurrent_quantizer":
            constraints.serialize(self.recurrent_quantizer_internal),
        "bias_quantizer":
            constraints.serialize(self.bias_quantizer_internal),
        "state_quantizer":
            constraints.serialize(self.state_quantizer_internal)
    }
    base_config = super(QLSTMCell, self).get_config()
    return dict(list(base_config.items()) + list(config.items()))


class QLSTM(RNN, PrunableLayer):
  """
  Class for the QLSTM layer.

  Most of these parameters follow the implementation of LSTM in
  Keras, with the exception of kernel_quantizer, recurrent_quantizer,
  bias_quantizer, state_quantizer.


  kernel_quantizer: quantizer function/class for kernel
  recurrent_quantizer: quantizer function/class for recurrent kernel
  bias_quantizer: quantizer function/class for bias
  state_quantizer: quantizer function/class for states

  We refer the reader to the documentation of LSTM in Keras for the
  other parameters.

  """

  def __init__(self,
               units,
               activation='quantized_tanh',
               recurrent_activation='hard_sigmoid',
               use_bias=True,
               kernel_initializer='glorot_uniform',
               recurrent_initializer='orthogonal',
               bias_initializer='zeros',
               unit_forget_bias=True,
               kernel_regularizer=None,
               recurrent_regularizer=None,
               bias_regularizer=None,
               activity_regularizer=None,
               kernel_constraint=None,
               recurrent_constraint=None,
               bias_constraint=None,
               kernel_quantizer=None,
               recurrent_quantizer=None,
               bias_quantizer=None,
               state_quantizer=None,
               dropout=0.,
               recurrent_dropout=0.,
               implementation=1,
               return_sequences=False,
               return_state=False,
               go_backwards=False,
               stateful=False,
               unroll=False,
               **kwargs):
    if implementation == 0:
      print('`implementation=0` has been deprecated, '
              'and now defaults to `implementation=1`.'
              'Please update your layer call.')

    if 'enable_caching_device' in kwargs:
      cell_kwargs = {'enable_caching_device':
                     kwargs.pop('enable_caching_device')}
    else:
      cell_kwargs = {}

    cell = QLSTMCell(
        units,
        activation=activation,
        recurrent_activation=recurrent_activation,
        use_bias=use_bias,
        kernel_initializer=kernel_initializer,
        recurrent_initializer=recurrent_initializer,
        unit_forget_bias=unit_forget_bias,
        bias_initializer=bias_initializer,
        kernel_regularizer=kernel_regularizer,
        recurrent_regularizer=recurrent_regularizer,
        bias_regularizer=bias_regularizer,
        kernel_constraint=kernel_constraint,
        recurrent_constraint=recurrent_constraint,
        bias_constraint=bias_constraint,
        kernel_quantizer=kernel_quantizer,
        recurrent_quantizer=recurrent_quantizer,
        bias_quantizer=bias_quantizer,
        state_quantizer=state_quantizer,
        dropout=dropout,
        recurrent_dropout=recurrent_dropout,
        implementation=implementation,
        dtype=kwargs.get('dtype'),
        trainable=kwargs.get('trainable', True),
        **cell_kwargs)

    super(QLSTM, self).__init__(
        cell,
        return_sequences=return_sequences,
        return_state=return_state,
        go_backwards=go_backwards,
        stateful=stateful,
        unroll=unroll,
        **kwargs)
    self.activity_regularizer = regularizers.get(activity_regularizer)
    self.input_spec = [tf.keras.layers.InputSpec(ndim=3)]

  def call(self, inputs, mask=None, training=None, initial_state=None):
    self._maybe_reset_cell_dropout_mask(self.cell)
    return super(QLSTM, self).call(
        inputs, mask=mask, training=training, initial_state=initial_state)

  def get_quantizers(self):
    return self.cell.quantizers

  def get_prunable_weights(self):
    return [self.cell.kernel, self.cell.recurrent_kernel]

  @property
  def units(self):
    return self.cell.units

  @property
  def activation(self):
    return self.cell.activation

  @property
  def recurrent_activation(self):
    return self.cell.recurrent_activation

  @property
  def use_bias(self):
    return self.cell.use_bias

  @property
  def kernel_initializer(self):
    return self.cell.kernel_initializer

  @property
  def recurrent_initializer(self):
    return self.cell.recurrent_initializer

  @property
  def bias_initializer(self):
    return self.cell.bias_initializer

  @property
  def unit_forget_bias(self):
    return self.cell.unit_forget_bias

  @property
  def kernel_regularizer(self):
    return self.cell.kernel_regularizer

  @property
  def recurrent_regularizer(self):
    return self.cell.recurrent_regularizer

  @property
  def bias_regularizer(self):
    return self.cell.bias_regularizer

  @property
  def kernel_constraint(self):
    return self.cell.kernel_constraint

  @property
  def recurrent_constraint(self):
    return self.cell.recurrent_constraint

  @property
  def bias_constraint(self):
    return self.cell.bias_constraint

  @property
  def kernel_quantizer_internal(self):
    return self.cell.kernel_quantizer_internal

  @property
  def recurrent_quantizer_internal(self):
    return self.cell.recurrent_quantizer_internal

  @property
  def bias_quantizer_internal(self):
    return self.cell.bias_quantizer_internal

  @property
  def state_quantizer_internal(self):
    return self.cell.state_quantizer_internal

  @property
  def kernel_quantizer(self):
    return self.cell.kernel_quantizer

  @property
  def recurrent_quantizer(self):
    return self.cell.recurrent_quantizer

  @property
  def bias_quantizer(self):
    return self.cell.bias_quantizer

  @property
  def state_quantizer(self):
    return self.cell.state_quantizer

  @property
  def dropout(self):
    return self.cell.dropout

  @property
  def recurrent_dropout(self):
    return self.cell.recurrent_dropout

  @property
  def implementation(self):
    return self.cell.implementation

  def get_config(self):
    config = {
        'units':
            self.units,
        'activation':
            activations.serialize(self.activation),
        'recurrent_activation':
            activations.serialize(self.recurrent_activation),
        'use_bias':
            self.use_bias,
        'kernel_initializer':
            initializers.serialize(self.kernel_initializer),
        'recurrent_initializer':
            initializers.serialize(self.recurrent_initializer),
        'bias_initializer':
            initializers.serialize(self.bias_initializer),
        'unit_forget_bias':
            self.unit_forget_bias,
        'kernel_regularizer':
            regularizers.serialize(self.kernel_regularizer),
        'recurrent_regularizer':
            regularizers.serialize(self.recurrent_regularizer),
        'bias_regularizer':
            regularizers.serialize(self.bias_regularizer),
        'activity_regularizer':
            regularizers.serialize(self.activity_regularizer),
        'kernel_constraint':
            constraints.serialize(self.kernel_constraint),
        'recurrent_constraint':
            constraints.serialize(self.recurrent_constraint),
        'bias_constraint':
            constraints.serialize(self.bias_constraint),
        "kernel_quantizer":
            constraints.serialize(self.kernel_quantizer_internal),
        "recurrent_quantizer":
            constraints.serialize(self.recurrent_quantizer_internal),
        "bias_quantizer":
            constraints.serialize(self.bias_quantizer_internal),
        "state_quantizer":
            constraints.serialize(self.state_quantizer_internal),
        'dropout':
            self.dropout,
        'recurrent_dropout':
            self.recurrent_dropout,
        'implementation':
            self.implementation
    }
    base_config = super(QLSTM, self).get_config()
    del base_config['cell']
    return dict(list(base_config.items()) + list(config.items()))

  def get_quantization_config(self):
    return {
        "kernel_quantizer":
            str(self.kernel_quantizer_internal),
        "recurrent_quantizer":
            str(self.recurrent_quantizer_internal),
        "bias_quantizer":
            str(self.bias_quantizer_internal),
        "state_quantizer":
            str(self.state_quantizer_internal),
        "activation":
            str(self.activation),
        "recurrent_activation":
            str(self.recurrent_activation),
    }

  @classmethod
  def from_config(cls, config):
    if 'implementation' in config and config['implementation'] == 0:
      config['implementation'] = 1
    return cls(**config)


class QGRUCell(GRUCell):
  """
  Cell class for the QGRUCell layer.

  Most of these parameters follow the implementation of GRUCell in
  Keras, with the exception of kernel_quantizer, recurrent_quantizer,
  bias_quantizer and state_quantizer.


  kernel_quantizer: quantizer function/class for kernel
  recurrent_quantizer: quantizer function/class for recurrent kernel
  bias_quantizer: quantizer function/class for bias
  state_quantizer: quantizer function/class for states


  We refer the reader to the documentation of GRUCell in Keras for the
  other parameters.

  """
  def __init__(self,
               units,
               activation='quantized_tanh',
               recurrent_activation='hard_sigmoid',
               use_bias=True,
               kernel_initializer='glorot_uniform',
               recurrent_initializer='orthogonal',
               bias_initializer='zeros',
               kernel_regularizer=None,
               recurrent_regularizer=None,
               bias_regularizer=None,
               kernel_constraint=None,
               recurrent_constraint=None,
               bias_constraint=None,
               kernel_quantizer=None,
               recurrent_quantizer=None,
               bias_quantizer=None,
               state_quantizer=None,
               dropout=0.,
               recurrent_dropout=0.,
               implementation=1,
               reset_after=False,
               **kwargs):

    self.kernel_quantizer = kernel_quantizer
    self.recurrent_quantizer = recurrent_quantizer
    self.bias_quantizer = bias_quantizer
    self.state_quantizer = state_quantizer

    self.kernel_quantizer_internal = get_quantizer(self.kernel_quantizer)
    self.recurrent_quantizer_internal = get_quantizer(self.recurrent_quantizer)
    self.bias_quantizer_internal = get_quantizer(self.bias_quantizer)
    self.state_quantizer_internal = get_quantizer(self.state_quantizer)

    self.quantizers = [
      self.kernel_quantizer_internal,
      self.recurrent_quantizer_internal,
      self.bias_quantizer_internal,
      self.state_quantizer_internal
    ]

    if hasattr(self.kernel_quantizer_internal, "_set_trainable_parameter"):
      self.kernel_quantizer_internal._set_trainable_parameter()

    if hasattr(self.recurrent_quantizer_internal, "_set_trainable_parameter"):
      self.recurrent_quantizer_internal._set_trainable_parameter()

    kernel_constraint, kernel_initializer = (
        get_auto_range_constraint_initializer(self.kernel_quantizer_internal,
                                              kernel_constraint,
                                              kernel_initializer))

    recurrent_constraint, recurrent_initializer = (
        get_auto_range_constraint_initializer(self.recurrent_quantizer_internal,
                                              recurrent_constraint,
                                              recurrent_initializer))

    if use_bias:
      bias_constraint, bias_initializer = (
          get_auto_range_constraint_initializer(self.bias_quantizer_internal,
                                                bias_constraint,
                                                bias_initializer))

    if activation is not None:
      activation = get_quantizer(activation)

    if recurrent_activation is not None:
      recurrent_activation = get_quantizer(recurrent_activation)

    super(QGRUCell, self).__init__(
      units=units,
      activation=activation,
      recurrent_activation=recurrent_activation,
      use_bias=use_bias,
      kernel_initializer=kernel_initializer,
      recurrent_initializer=recurrent_initializer,
      bias_initializer=bias_initializer,
      kernel_regularizer=kernel_regularizer,
      recurrent_regularizer=recurrent_regularizer,
      bias_regularizer=bias_regularizer,
      kernel_constraint=kernel_constraint,
      recurrent_constraint=recurrent_constraint,
      bias_constraint=bias_constraint,
      dropout=dropout,
      recurrent_dropout=recurrent_dropout,
      implementation=implementation,
      reset_after=reset_after,
      **kwargs)

  def call(self, inputs, states, training=None):
    # previous memory
    h_tm1_tmp = states[0] if nest.is_sequence(states) else states

    dp_mask = self.get_dropout_mask_for_cell(inputs, training, count=3)
    rec_dp_mask = self.get_recurrent_dropout_mask_for_cell(
        h_tm1_tmp, training, count=3)

    if self.state_quantizer:
      h_tm1 = self.state_quantizer_internal(h_tm1_tmp)
    else:
      h_tm1 = h_tm1_tmp

    if self.kernel_quantizer:
      quantized_kernel = self.kernel_quantizer_internal(self.kernel)
    else:
      quantized_kernel = self.kernel
    if self.recurrent_quantizer:
      quantized_recurrent = self.recurrent_quantizer_internal(self.recurrent_kernel)
    else:
      quantized_recurrent = self.kernel

    if self.use_bias:
      if self.bias_quantizer:
        quantized_bias = self.bias_quantizer_internal(self.bias)
      else:
        quantized_bias = self.bias

      if not self.reset_after:
        input_bias, recurrent_bias = quantized_bias, None
      else:
        input_bias, recurrent_bias = array_ops.unstack(quantized_bias)

    if self.implementation == 1:
      if 0. < self.dropout < 1.:
        inputs_z = inputs * dp_mask[0]
        inputs_r = inputs * dp_mask[1]
        inputs_h = inputs * dp_mask[2]
      else:
        inputs_z = inputs
        inputs_r = inputs
        inputs_h = inputs

      x_z = K.dot(inputs_z, quantized_kernel[:, :self.units])
      x_r = K.dot(inputs_r, quantized_kernel[:, self.units:self.units * 2])
      x_h = K.dot(inputs_h, quantized_kernel[:, self.units * 2:])

      if self.use_bias:
        x_z = K.bias_add(x_z, input_bias[:self.units])
        x_r = K.bias_add(x_r, input_bias[self.units: self.units * 2])
        x_h = K.bias_add(x_h, input_bias[self.units * 2:])

      if 0. < self.recurrent_dropout < 1.:
        h_tm1_z = h_tm1 * rec_dp_mask[0]
        h_tm1_r = h_tm1 * rec_dp_mask[1]
        h_tm1_h = h_tm1 * rec_dp_mask[2]
      else:
        h_tm1_z = h_tm1
        h_tm1_r = h_tm1
        h_tm1_h = h_tm1

      recurrent_z = K.dot(h_tm1_z, quantized_recurrent[:, :self.units])
      recurrent_r = K.dot(h_tm1_r,
                          quantized_recurrent[:, self.units:self.units * 2])
      if self.reset_after and self.use_bias:
        recurrent_z = K.bias_add(recurrent_z, recurrent_bias[:self.units])
        recurrent_r = K.bias_add(recurrent_r,
                                 recurrent_bias[self.units:self.units * 2])

      z = self.recurrent_activation(x_z + recurrent_z)
      r = self.recurrent_activation(x_r + recurrent_r)

      # reset gate applied after/before matrix multiplication
      if self.reset_after:
        recurrent_h = K.dot(h_tm1_h, quantized_recurrent[:, self.units * 2:])
        if self.use_bias:
          recurrent_h = K.bias_add(recurrent_h, recurrent_bias[self.units * 2:])
        recurrent_h = r * recurrent_h
      else:
        recurrent_h = K.dot(r * h_tm1_h,
                            quantized_recurrent[:, self.units * 2:])

      hh = self.activation(x_h + recurrent_h)
    else:
      if 0. < self.dropout < 1.:
        inputs = inputs * dp_mask[0]

      # inputs projected by all gate matrices at once
      matrix_x = K.dot(inputs, quantized_kernel)
      if self.use_bias:
        # biases: bias_z_i, bias_r_i, bias_h_i
        matrix_x = K.bias_add(matrix_x, input_bias)

      x_z, x_r, x_h = array_ops.split(matrix_x, 3, axis=-1)

      if self.reset_after:
        # hidden state projected by all gate matrices at once
        matrix_inner = K.dot(h_tm1, quantized_recurrent)
        if self.use_bias:
          matrix_inner = K.bias_add(matrix_inner, recurrent_bias)
      else:
        # hidden state projected separately for update/reset and new
        matrix_inner = K.dot(h_tm1, quantized_recurrent[:, :2 * self.units])

      recurrent_z, recurrent_r, recurrent_h = array_ops.split(
          matrix_inner, [self.units, self.units, -1], axis=-1)

      z = self.recurrent_activation(x_z + recurrent_z)
      r = self.recurrent_activation(x_r + recurrent_r)

      if self.reset_after:
        recurrent_h = r * recurrent_h
      else:
        recurrent_h = K.dot(r * h_tm1,
                            quantized_recurrent[:, 2 * self.units:])

      hh = self.activation(x_h + recurrent_h)
    # previous and candidate state mixed by update gate
    h = z * h_tm1 + (1 - z) * hh
    return h, [h]

  def get_config(self):
    config = {
        "kernel_quantizer":
            constraints.serialize(self.kernel_quantizer_internal),
        "recurrent_quantizer":
            constraints.serialize(self.recurrent_quantizer_internal),
        "bias_quantizer":
            constraints.serialize(self.bias_quantizer_internal),
        "state_quantizer":
            constraints.serialize(self.state_quantizer_internal)
    }
    base_config = super(QGRUCell, self).get_config()
    return dict(list(base_config.items()) + list(config.items()))


class QGRU(RNN, PrunableLayer):
  """
  Class for the QGRU layer.

  Most of these parameters follow the implementation of GRU in
  Keras, with the exception of kernel_quantizer, recurrent_quantizer,
  bias_quantizer and state_quantizer.


  kernel_quantizer: quantizer function/class for kernel
  recurrent_quantizer: quantizer function/class for recurrent kernel
  bias_quantizer: quantizer function/class for bias
  state_quantizer: quantizer function/class for states


  We refer the reader to the documentation of GRU in Keras for the
  other parameters.

  """

  def __init__(self,
               units,
               activation='quantized_tanh',
               recurrent_activation='hard_sigmoid',
               use_bias=True,
               kernel_initializer='glorot_uniform',
               recurrent_initializer='orthogonal',
               bias_initializer='zeros',
               kernel_regularizer=None,
               recurrent_regularizer=None,
               bias_regularizer=None,
               activity_regularizer=None,
               kernel_constraint=None,
               recurrent_constraint=None,
               bias_constraint=None,
               kernel_quantizer=None,
               recurrent_quantizer=None,
               bias_quantizer=None,
               state_quantizer=None,
               dropout=0.,
               recurrent_dropout=0.,
               implementation=1,
               return_sequences=False,
               return_state=False,
               go_backwards=False,
               stateful=False,
               unroll=False,
               reset_after=False,
               **kwargs):
    if implementation == 0:
      print('`implementation=0` has been deprecated, '
              'and now defaults to `implementation=1`.'
              'Please update your layer call.')

    if 'enable_caching_device' in kwargs:
      cell_kwargs = {'enable_caching_device':
                     kwargs.pop('enable_caching_device')}
    else:
      cell_kwargs = {}

    cell = QGRUCell(
        units,
        activation=activation,
        recurrent_activation=recurrent_activation,
        use_bias=use_bias,
        kernel_initializer=kernel_initializer,
        recurrent_initializer=recurrent_initializer,
        bias_initializer=bias_initializer,
        kernel_regularizer=kernel_regularizer,
        recurrent_regularizer=recurrent_regularizer,
        bias_regularizer=bias_regularizer,
        kernel_constraint=kernel_constraint,
        recurrent_constraint=recurrent_constraint,
        bias_constraint=bias_constraint,
        kernel_quantizer=kernel_quantizer,
        recurrent_quantizer=recurrent_quantizer,
        bias_quantizer=bias_quantizer,
        state_quantizer=state_quantizer,
        dropout=dropout,
        recurrent_dropout=recurrent_dropout,
        implementation=implementation,
        reset_after=reset_after,
        dtype=kwargs.get('dtype'),
        trainable=kwargs.get('trainable', True),
        **cell_kwargs)

    super(QGRU, self).__init__(
        cell,
        return_sequences=return_sequences,
        return_state=return_state,
        go_backwards=go_backwards,
        stateful=stateful,
        unroll=unroll,
        **kwargs)
    self.activity_regularizer = regularizers.get(activity_regularizer)
    self.input_spec = [tf.keras.layers.InputSpec(ndim=3)]

  def call(self, inputs, mask=None, training=None, initial_state=None):
    self._maybe_reset_cell_dropout_mask(self.cell)
    return super(QGRU, self).call(
        inputs, mask=mask, training=training, initial_state=initial_state)

  def get_quantizers(self):
    return self.cell.quantizers

  def get_prunable_weights(self):
    return [self.cell.kernel, self.cell.recurrent_kernel]

  @property
  def units(self):
    return self.cell.units

  @property
  def activation(self):
    return self.cell.activation

  @property
  def recurrent_activation(self):
    return self.cell.recurrent_activation

  @property
  def use_bias(self):
    return self.cell.use_bias

  @property
  def kernel_initializer(self):
    return self.cell.kernel_initializer

  @property
  def recurrent_initializer(self):
    return self.cell.recurrent_initializer

  @property
  def bias_initializer(self):
    return self.cell.bias_initializer

  @property
  def kernel_regularizer(self):
    return self.cell.kernel_regularizer

  @property
  def recurrent_regularizer(self):
    return self.cell.recurrent_regularizer

  @property
  def bias_regularizer(self):
    return self.cell.bias_regularizer

  @property
  def kernel_constraint(self):
    return self.cell.kernel_constraint

  @property
  def recurrent_constraint(self):
    return self.cell.recurrent_constraint

  @property
  def bias_constraint(self):
    return self.cell.bias_constraint

  @property
  def kernel_quantizer_internal(self):
    return self.cell.kernel_quantizer_internal

  @property
  def recurrent_quantizer_internal(self):
    return self.cell.recurrent_quantizer_internal

  @property
  def bias_quantizer_internal(self):
    return self.cell.bias_quantizer_internal

  @property
  def state_quantizer_internal(self):
    return self.cell.state_quantizer_internal

  @property
  def kernel_quantizer(self):
    return self.cell.kernel_quantizer

  @property
  def recurrent_quantizer(self):
    return self.cell.recurrent_quantizer

  @property
  def bias_quantizer(self):
    return self.cell.bias_quantizer

  @property
  def state_quantizer(self):
    return self.cell.state_quantizer

  @property
  def dropout(self):
    return self.cell.dropout

  @property
  def recurrent_dropout(self):
    return self.cell.recurrent_dropout

  @property
  def implementation(self):
    return self.cell.implementation

  @property
  def reset_after(self):
    return self.cell.reset_after

  def get_config(self):
    config = {
        'units':
            self.units,
        'activation':
            activations.serialize(self.activation),
        'recurrent_activation':
            activations.serialize(self.recurrent_activation),
        'use_bias':
            self.use_bias,
        'kernel_initializer':
            initializers.serialize(self.kernel_initializer),
        'recurrent_initializer':
            initializers.serialize(self.recurrent_initializer),
        'bias_initializer':
            initializers.serialize(self.bias_initializer),
        'kernel_regularizer':
            regularizers.serialize(self.kernel_regularizer),
        'recurrent_regularizer':
            regularizers.serialize(self.recurrent_regularizer),
        'bias_regularizer':
            regularizers.serialize(self.bias_regularizer),
        'activity_regularizer':
            regularizers.serialize(self.activity_regularizer),
        'kernel_constraint':
            constraints.serialize(self.kernel_constraint),
        'recurrent_constraint':
            constraints.serialize(self.recurrent_constraint),
        'bias_constraint':
            constraints.serialize(self.bias_constraint),
        "kernel_quantizer":
            constraints.serialize(self.kernel_quantizer_internal),
        "recurrent_quantizer":
            constraints.serialize(self.recurrent_quantizer_internal),
        "bias_quantizer":
            constraints.serialize(self.bias_quantizer_internal),
        "state_quantizer":
            constraints.serialize(self.state_quantizer_internal),
        'dropout':
            self.dropout,
        'recurrent_dropout':
            self.recurrent_dropout,
        'implementation':
            self.implementation,
        'reset_after':
            self.reset_after
    }
    base_config = super(QGRU, self).get_config()
    del base_config['cell']
    return dict(list(base_config.items()) + list(config.items()))

  def get_quantization_config(self):
    return {
        "kernel_quantizer":
            str(self.kernel_quantizer_internal),
        "recurrent_quantizer":
            str(self.recurrent_quantizer_internal),
        "bias_quantizer":
            str(self.bias_quantizer_internal),
        "state_quantizer":
            str(self.state_quantizer_internal),
        "activation":
            str(self.activation),
        "recurrent_activation":
            str(self.recurrent_activation),
    }

  @classmethod
  def from_config(cls, config):
    if 'implementation' in config and config['implementation'] == 0:
      config['implementation'] = 1
    return cls(**config)


class QBidirectional(Bidirectional):
  """
  Class for the QBidirecitonal wrapper.

  Most of these parameters follow the implementation of Bidirectional in
  Keras.

  We refer the reader to the documentation of Bidirectional in Keras for the
  other parameters.

  """
  def get_quantizers(self):
    """
    Returns quantizers in the order they were created.
    """
    return self.forward_layer.get_quantizers() + self.backward_layer.get_quantizers()

  @property
  def activation(self):
    return self.layer.activation

  def get_quantization_config(self):
    return {
      "layer" : self.layer.get_quantization_config(),
      "backward_layer" : self.backward_layer.get_quantization_config()
    }



# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import numpy as np
from tensorflow.keras import constraints

import tensorflow as tf
import tensorflow.keras.backend as K

from tensorflow.keras.layers import AveragePooling2D
from tensorflow.keras.layers import GlobalAveragePooling2D
from .qlayers import QActivation
from .quantizers import get_quantizer


class QAveragePooling2D(AveragePooling2D):
  """Computes the quantized version of AveragePooling2D."""

  def __init__(self, pool_size=(2, 2),
               strides=None,
               padding="valid",
               data_format=None,
               average_quantizer=None,
               activation=None,
               **kwargs):

    self.average_quantizer = average_quantizer
    self.average_quantizer_internal = get_quantizer(self.average_quantizer)
    self.quantizers = [self.average_quantizer_internal]

    if activation is not None:
      self.activation = get_quantizer(activation)
    else:
      self.activation = activation

    super(QAveragePooling2D, self).__init__(
        pool_size=pool_size, strides=strides, padding=padding,
        data_format=data_format, **kwargs)

  def call(self, inputs):
    """Performs quantized AveragePooling followed by QActivation.

    Since there is no specific parameter for averaging op, we couldn't apply
    averaging quantizer to the averaging op. We have two options:
    1. we perform our own average as sum first then multiply with the
       inversion
       of the division factor: sum(x) * quantize(1/pool_area)
    2. first, we call keras version of averaging first: y1 = keras_average(x)
       then multiply it with pool_size^2: y2 = y1 * pool_area
       Last, y3 = y2 * quantize(1/ pool_area)
    Our numerical anaysis suggests negligible error between 1 and 2. Therefore
    we use option #2 here for the simplicity of implementation.

    """

    x = super(QAveragePooling2D, self).call(inputs)

    if self.average_quantizer:
      if isinstance(self.pool_size, int):
        pool_area = self.pool_size * self.pool_size
      else:
        pool_area = np.prod(self.pool_size)

      # Revertes the division results.
      x = x * pool_area

      # Quantizes the multiplication factor.
      mult_factor = 1.0 / pool_area

      q_mult_factor = self.average_quantizer_internal(mult_factor)
      q_mult_factor = K.cast_to_floatx(q_mult_factor)
      x = x * q_mult_factor

    if self.activation is not None:
      return self.activation(x)
    return x

  def get_config(self):
    config = {
        "average_quantizer":
            constraints.serialize(self.average_quantizer_internal),
        "activation": constraints.serialize(self.activation),
    }
    base_config = super(QAveragePooling2D, self).get_config()
    return dict(list(base_config.items()) + list(config.items()))

  def get_quantization_config(self):
    return {
        "average_quantizer":
            str(self.average_quantizer_internal),
        "activation":
            str(self.activation)
    }

  def get_quantizers(self):
    return self.quantizers


class QGlobalAveragePooling2D(GlobalAveragePooling2D):
  """Computes the quantized version of GlobalAveragePooling2D."""

  def __init__(self, data_format=None,
               average_quantizer=None,
               activation=None,
               **kwargs):

    self.average_quantizer = average_quantizer
    self.average_quantizer_internal = get_quantizer(self.average_quantizer)
    self.quantizers = [self.average_quantizer_internal]

    if activation is not None:
      self.activation = get_quantizer(activation)
    else:
      self.activation = activation

    super(QGlobalAveragePooling2D, self).__init__(
        data_format=data_format, **kwargs)

  def compute_pooling_area(self, input_shape):
    input_shape = input_shape.as_list()
    if self.data_format == "channels_last":
      return input_shape[1] * input_shape[2]
    else:
      return input_shape[2] * input_shape[3]

  def call(self, inputs):
    """Performs quantized GlobalAveragePooling followed by QActivation.

    Since there is no specific parameter for averaging op, we couldn't apply
    averaging quantizer to the averaging op. We have two options:
    1. we perform our own average as sum first then multiply with the
       inversion
       of the division factor: sum(x) * quantize(1/pool_area)
    2. first, we call keras version of averaging first:
       y1 = keras_global_average(x)
       then multiply it with the denominator(pool_area) used by averaging:
       y2 = y1 * pool_area
       Last, y3 = y2 * quantize(1/ pool_area)
    Our numerical anaysis suggests negligible error between 1 and 2. Therefore
    we use option #2 here for the simplicity of implementation.

    """

    x = super(QGlobalAveragePooling2D, self).call(inputs)

    if self.average_quantizer:
      pool_area = self.compute_pooling_area(input_shape=inputs.shape)
      # Reverts the division results
      x = x * pool_area
      # Quantizes the multiplication factor
      mult_factor = 1.0 / pool_area
      q_mult_factor = self.average_quantizer_internal(mult_factor)
      x = x * q_mult_factor

    if self.activation is not None:
      return self.activation(x)
    return x

  def get_config(self):
    config = {
        "average_quantizer":
            constraints.serialize(self.average_quantizer_internal),
        "activation": constraints.serialize(self.activation),
    }
    base_config = super(QGlobalAveragePooling2D, self).get_config()
    return dict(list(base_config.items()) + list(config.items()))

  def get_quantization_config(self):
    return {
        "average_quantizer":
            str(self.average_quantizer_internal),
        "activation":
            str(self.activation)
    }

  def get_quantizers(self):
    return self.quantizers



# Lint as: python3
# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Calculate energy consumption of a given quantized model."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

from qkeras.qtools.generate_layer_data_type_map import KERAS_LAYERS
from qkeras.qtools.generate_layer_data_type_map import QKERAS_LAYERS
from qkeras.qtools.quantized_operators.quantizer_impl import IQuantizer
from qkeras.qtools.settings import cfg

# Model based on：
#   Mark Horowitz, Computing’s Energy Problem (and what we can
#   do about it). IEEE ISSCC, pp. 10–14, 2014
#   www.youtube.com/watch?v=eZdOkDtYMoo&feature=youtu.be&t=497

# all metrics converted to pJ/bit

OP = {
    "fp32": {
        "add": lambda x: max(cfg.fp32_add(x), 0),
        "mul": lambda x: max(cfg.fp32_mul(x), 0)
    },
    "fp16": {
        "add": lambda x: max(cfg.fp16_add(x), 0),
        "mul": lambda x: max(cfg.fp16_mul(x), 0)
    },
    "fpm": {
        "add": lambda x: max(cfg.fpm_add(x), 0),
        "mux": lambda x: max(cfg.fpm_add(x), 0),
        "xor": lambda x: max(cfg.fpm_add(x), 0),
        "and": lambda x: max(cfg.fpm_add(x), 0),
        "or": lambda x: max(cfg.fpm_add(x), 0),
        "shifter": lambda x: max(cfg.fpm_add(x), 0),
        "mul": lambda x: max(cfg.fpm_mul(x), 0)
    },
    "sram": {"rd": lambda x: max(cfg.sram_rd(x), 0),
             "wr": lambda x: max(cfg.sram_rd(x), 0),
             "mul_factor": cfg.sram_mul_factor},
    "dram": {"rd": lambda x: max(cfg.dram_rd(x), 0),
             "wr": lambda x: max(cfg.dram_rd(x), 0),
             "mul_factor": cfg.dram_mul_factor}
}


def get_op_type(quantizer):
  assert isinstance(quantizer, IQuantizer)

  if quantizer.is_floating_point:
    return "fp" + str(quantizer.bits)
  else:
    return "fpm"


def memory_read_energy(is_input_layer, tensor_shape, mode, min_sram_size,
                       rd_wr_on_io, quantizer_bits, is_tensor=True):
  """compute energy to bring tensors from DRAM to SRAM."""

  if is_input_layer:
    if rd_wr_on_io:
      mode = "dram"
    else:
      mode = "sram"

  energy_mem = 0

  if is_tensor:
    tensor_shape = tensor_shape[1:]

  total_bits = np.prod(tensor_shape) * quantizer_bits
  total_bits_log2 = np.log2(max(total_bits, min_sram_size))

  if mode == "dram":
    # load input from dram; wx_sizes[1]-> input x quantizer bits
    # total_bits * 20
    energy_mem += OP["dram"]["rd"](total_bits)
    if rd_wr_on_io:
      # write input to sram
      # total_bits * sqrt(data_size/2^18)*0.3125
      # bits1 = total_bits * OP["sram"]["mul_factor"](np.prod(tensor_shape))
      # energy_mem += OP["sram"]["wr"](bits1)
      energy_mem += (
          np.ceil(total_bits * OP["sram"]["mul_factor"]) *
          OP["sram"]["wr"](total_bits_log2)
      )
  elif mode == "sram":
    # read input from sram
    # total_bits * sqrt(data_size/2^18)*0.3125
    # bits1 = total_bits * OP["sram"]["mul_factor"](np.prod(tensor_shape))
    # energy_mem += OP["sram"]["rd"](bits1)
    energy_mem += (
        np.ceil(total_bits * OP["sram"]["mul_factor"]) *
        OP["sram"]["rd"](total_bits_log2)
    )

  return energy_mem


def parameter_read_energy(
    layer, layer_item, weights_on_memory, min_sram_size, rd_wr_on_io):
  """read weights/bias from memory."""

  node_type = layer.__class__.__name__
  rd_energy = 0
  if node_type in ["QBatchNormalization", "BatchNormalization"]:
    gamma_quantizer = layer_item["gamma_quantizer"]
    beta_quantizer = layer_item["beta_quantizer"]
    mean_quantizer = layer_item["mean_quantizer"]
    variance_quantizer = layer_item["variance_quantizer"]

    # gamma, beta, mean, stddev
    weights = layer.get_weights()
    s = len(weights[0])
    for q in [gamma_quantizer, beta_quantizer, mean_quantizer,
              variance_quantizer]:
      if q:
        rd_energy += memory_read_energy(
            False, (s), weights_on_memory, min_sram_size, rd_wr_on_io,
            q.bits, is_tensor=False)

  elif node_type in QKERAS_LAYERS or node_type in KERAS_LAYERS:
    weight_quantizer = layer_item.weight_quantizer
    w_shapes = layer_item.w_shapes
    bias_quantizer = layer_item.bias_quantizer
    b_shapes = layer_item.b_shapes

    rd_energy += memory_read_energy(
        False, w_shapes, weights_on_memory, min_sram_size, rd_wr_on_io,
        weight_quantizer.bits, is_tensor=False
    )

    if bias_quantizer:
      # if use_bias=0, no bias
      bias_shapes = (b_shapes)
      rd_energy += memory_read_energy(
          False, bias_shapes, weights_on_memory, min_sram_size, rd_wr_on_io,
          bias_quantizer.bits, is_tensor=False
      )

  return rd_energy


def memory_write_energy(is_output_layer, tensor_shape, mode, min_sram_size,
                        rd_wr_on_io, quantizer_bits):
  """compute energy to bring tensors from SRAM to DRAM."""
  if is_output_layer:
    if rd_wr_on_io:
      mode = "dram"
    else:
      mode = "sram"

  energy_mem = 0

  tensor_shape = tensor_shape[1:]

  total_bits = np.prod(tensor_shape) * quantizer_bits
  total_bits_log2 = np.log2(max(total_bits, min_sram_size))

  if mode == "dram":
    # load input from dram; wx_sizes[1]-> input x quantizer bits
    if rd_wr_on_io:
      # read input from sram
      # total_bits * sqrt(data_size/2^18)*0.3125
      # bits1 = total_bits * OP["sram"]["mul_factor"](np.prod(tensor_shape))
      # energy_mem += OP["sram"]["rd"](bits1)
      energy_mem += (
          np.ceil(total_bits * OP["sram"]["mul_factor"]) *
          OP["sram"]["rd"](total_bits_log2)
      )
    # write output to dram
    energy_mem += OP["dram"]["wr"](total_bits)

  elif mode == "sram":
    # write to sram
    # total_bits * sqrt(data_size/2^18)*0.3125
    # bits1 = total_bits * OP["sram"]["mul_factor"](np.prod(tensor_shape))
    # energy_mem +=  OP["sram"]["wr"](bits1)
    energy_mem += (
        np.ceil(total_bits * OP["sram"]["mul_factor"]) *
        OP["sram"]["wr"](total_bits_log2)
    )

  return energy_mem


def energy_estimate(model, layer_map, weights_on_memory,
                    activations_on_memory, min_sram_size,
                    rd_wr_on_io):
  """estimate energy."""

  output_layers = layer_map["output_layers"]
  input_layers = layer_map["input_layers"]
  layer_data_type_map = layer_map["layer_data_type_map"]

  result = {}
  total_energy = 0

  # compute MAC and memory access energy for intermediate layers
  for layer in model.layers:
    if layer not in layer_data_type_map.keys():
      continue

    layer_item = layer_data_type_map[layer]

    if hasattr(layer_item, "input_quantizer_list"):
      input_quantizer_list = layer_item.input_quantizer_list
      operation_count = layer_item.operation_count
      output_shapes = layer_item.output_shapes
      output_quantizer = layer_item.output_quantizer
    else:
      input_quantizer_list = layer_item["input_quantizer_list"]
      operation_count = layer_item["operation_count"]
      output_shapes = layer_item["output_shapes"]
      output_quantizer = layer_item["output_quantizer"]

    is_input_layer = layer in input_layers
    is_output_layer = layer in output_layers

    input_rd_energy = 0
    energy_op = 0
    input_shape = layer.input_shape
    if not isinstance(input_shape, list):
      input_shape = [input_shape]

    for (input_shape, input_quantizer) in zip(
        input_shape, input_quantizer_list):
      input_rd_energy += memory_read_energy(
          is_input_layer, input_shape,
          activations_on_memory, min_sram_size, rd_wr_on_io,
          input_quantizer.bits)

    parameter_rd_energy = parameter_read_energy(
        layer, layer_item, weights_on_memory, min_sram_size,
        rd_wr_on_io)

    output_wr_energy = memory_write_energy(
        is_output_layer, output_shapes,
        activations_on_memory, min_sram_size, rd_wr_on_io,
        output_quantizer.bits)
    # QActivation Layer
    if layer.__class__.__name__ in ["QActivation", "QAdaptiveActivation", "Activation"]:
      pass

    # QBN Layer
    elif layer.__class__.__name__ in [
        "QBatchNormalization", "BatchNormalization"]:
      # assume QBN is embedded with conv/dense layers
      # -> no memory read/write cost

      divider = layer_item["internal_divide_quantizer"]
      if divider:
        gate_factor = divider.gate_factor
        mode = divider.implemented_as()
        energy_op += gate_factor * OP[
            get_op_type(divider.output)][mode](divider.gate_bits)

      multiplier = layer_item["internal_multiplier"]
      if multiplier:
        gate_factor = multiplier.gate_factor
        mode = multiplier.implemented_as()
        energy_op += gate_factor * OP[
            get_op_type(multiplier.output)][mode](multiplier.gate_bits)

      energy_op *= operation_count

    # Merge layer
    elif layer.__class__.__name__ in ["Add", "Multiply", "Subtract"]:

      # multiply or add operation energy
      # TODO(lishanok): check energy for concatenate
      merge_quantizer = layer_item.multiplier
      mode = merge_quantizer.implemented_as()
      number_of_inputs = len(layer_item.input_quantizer_list)
      gate_factor = merge_quantizer.gate_factor

      q = get_op_type(merge_quantizer.output)
      b = merge_quantizer.gate_bits
      energy_op = (number_of_inputs - 1) * operation_count * gate_factor * OP[
          q][mode](b)

    # AveragePooling and GlobalAveragePooling
    elif layer.__class__.__name__ in [
        "AveragePooling2D", "AvgPool2D", "GlobalAvgPool2D",
        "GlobalAveragePooling2D"]:
      # accumulation operation energy
      accumulator = layer_item.accumulator
      add_energy = OP[get_op_type(accumulator.output)]["add"](
          accumulator.output.bits)
      energy_op = operation_count * add_energy

    # MAC energy calculation
    elif layer.__class__.__name__ in ["QConv2D", "QConv1D", "QDepthwiseConv2D",
                                      "QDense", "Conv2D", "Conv1D",
                                      "DepthwiseConv2D", "Dense"]:
      multiplier = layer_item.multiplier
      accumulator = layer_item.accumulator

      # implementation mode: xor/andgate/shift etc.
      mode = multiplier.implemented_as()
      gate_factor = multiplier.gate_factor

      op = get_op_type(multiplier.output)
      bits = multiplier.gate_bits
      c1 = gate_factor * OP[op][mode](bits)
      c2 = OP[get_op_type(accumulator.output)]["add"](accumulator.output.bits)
      energy_op = operation_count * (c1 + c2)

    else:
      pass

    result[layer.name] = {
        "class_name": layer.__class__.__name__,
        "energy": {
            "inputs": float("{0:.2f}".format(input_rd_energy)),
            "outputs": float("{0:.2f}".format(output_wr_energy)),
            "parameters": float("{0:.2f}".format(parameter_rd_energy)),
            "op_cost": float("{0:.2f}".format(energy_op))
        }
    }
    total_energy += input_rd_energy + output_wr_energy + parameter_rd_energy + energy_op

  result["total_cost"] = int(total_energy)

  return result



# Lint as: python3
# ==============================================================================
# Copyright 2020 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Implements forgiving factor metrics bit model size in bits."""

import numpy as np
import six
from qkeras.autoqkeras.forgiving_metrics.forgiving_factor import ForgivingFactor   # pylint: disable=line-too-long
from qkeras import get_quantizer


class ForgivingFactorBits(ForgivingFactor):
  """Implements forgiving factor with target as number of bits."""

  def __init__(
      self, delta_p, delta_n, rate, stress=1.0,
      input_bits=8, output_bits=8, ref_bits=8, config=None):
    self.stress = stress
    self.input_bits = input_bits
    self.output_bits = output_bits
    self.ref_bits = ref_bits
    self.ref_size = {}
    self.config = config if config else {}

    super(ForgivingFactorBits, self).__init__(delta_p, delta_n, rate)

  def _param_size(self, layer):
    """Computes size of parameters of a layer in bits."""
    t_size = self.ref_bits
    parameter_size = 0
    # we only compute parameter sizes for these layers, and BatchNormalization
    # is a special case because it exports mean and beta that is absorbed by
    # previous or next layer. As mean and beta will be compressed into a single
    # value, we actually only need to take care of the shape.
    if layer.__class__.__name__ in [
        "Dense", "Conv2D", "Conv1D", "DepthwiseConv2D"]:
      for w in layer.get_weights():
        parameter_size += t_size * np.prod(w.shape)
    elif layer.__class__.__name__ in [
        "QDense", "QConv2D", "QConv1D", "QDepthwiseConv2D"]:
      for i, w in enumerate(layer.get_weights()):
        if layer.get_quantizers()[i]:
          bits = layer.get_quantizers()[i].bits
        else:
          bits = t_size
        parameter_size += bits * np.prod(w.shape)
    elif layer.__class__.__name__ in ["BatchNormalization"]:
      # scale
      index = -1
      parameter_size += t_size * np.prod(layer.get_weights()[index].shape)
      # center (bias)
      if layer.center:
        index = int(bool(layer.scale))
        parameter_size += t_size * np.prod(layer.get_weights()[index].shape)
    elif layer.__class__.__name__ in ["QBatchNormalization"]:
      # scale
      index = -1
      bits = 6
      parameter_size += bits * np.prod(layer.get_weights()[index].shape)
      # center (bias)
      if layer.center:
        bits = 5
        index = int(bool(layer.scale))
        parameter_size += bits * np.prod(layer.get_weights()[index].shape)
    return parameter_size

  def _act_size(self, layer):
    """Computes size of activations of a layer in bits."""
    i_size = self.input_bits
    o_size = self.output_bits
    t_size = self.ref_bits
    output_size = np.prod(layer.output.shape[1:])
    # we compute activation sizes for inputs and outputs
    if layer.__class__.__name__ in ["InputLayer"]:
      return i_size * output_size
    elif layer.__class__.__name__ in [
        "Dense", "Conv2D", "Conv1D", "DepthwiseConv2D"]:
      if layer.activation is not None and layer.activation.__name__ != "linear":
        return t_size * output_size
      else:
        return 0
    elif layer.__class__.__name__ in [
        "QDense", "QConv2D", "QConv1D", "QDepthwiseConv2D"]:
      if layer.activation is None:
        is_softmax = False
        is_linear = False
      else:
        if isinstance(layer.activation, six.string_types):
          is_softmax = layer.activation == "softmax"
          is_linear = layer.activation == "linear"
        elif hasattr(layer.activation, "__name__"):
          is_softmax = layer.activation.__name__ == "softmax"
          is_linear = layer.activation.__name__ == "linear"
        else:
          is_softmax = False
          is_linear = False

        if is_softmax:
          bits = o_size
        elif is_linear:
          bits = 0
        else:
          assert not isinstance(layer.activation, six.string_types)
          if hasattr(layer.activation, "bits"):
            bits = layer.activation.bits
          else:
            bits = t_size

        return bits * np.prod(layer.output.shape.as_list()[1:])
    elif layer.__class__.__name__ in ["QActivation", "Activation"]:
      if isinstance(layer.activation, six.string_types):
        is_linear = layer.activation == "linear"
        is_softmax = layer.activation == "softmax"
        is_sigmoid = layer.activation == "sigmoid"
      else:
        is_linear = layer.activation.__name__ == "linear"
        is_softmax = layer.activation.__name__ == "softmax"
        is_sigmoid = layer.activation.__name__ == "sigmoid"

      if is_linear:
        bits = 0
      elif is_softmax or is_sigmoid:
        bits = o_size
      else:
        if isinstance(layer.activation, six.string_types):
          activation = get_quantizer(layer.activation)
        else:
          activation = layer.activation
        if hasattr(activation, "bits"):
          bits = activation.bits
        else:
          bits = t_size
      return bits * output_size
    return 0

  def compute_model_size(self, model):
    """Computes size of model."""

    a_size = 0
    p_size = 0
    total_size = 0
    model_size_dict = {}
    for layer in model.layers:
      layer_name = layer.__class__.__name__
      layer_config = self.config.get(
          layer_name, self.config.get("default", None))
      if layer_config:
        parameters = self._param_size(layer)
        activations = self._act_size(layer)
        p_weight = ("parameters" in layer_config)
        a_weight = ("activations" in layer_config)
        total = p_weight * parameters + a_weight * activations
        model_size_dict[layer.name] = {
            "parameters": parameters,
            "activations": activations,
            "total": total
        }
        a_size += a_weight * activations
        p_size += p_weight * parameters
        total_size += total

    return (total_size, p_size, a_size, model_size_dict)

  def get_reference(self, model):
    if not hasattr(self, "reference_size"):
      cached_result = self.compute_model_size(model)
      self.reference_size = cached_result[0] * self.stress
      self.ref_p = cached_result[1]
      self.ref_a = cached_result[2]
      self.reference_size_dict = cached_result[3]

    return self.reference_size

  def get_reference_stats(self):
    return self.reference_size_dict

  def get_trial(self, model):
    """Computes size of quantization trial."""

    result = self.compute_model_size(model)
    self.trial_size = result[0]
    self.total_p_bits = result[1]
    self.total_a_bits = result[2]
    self.trial_size_dict = result[3]

    return self.trial_size

  def get_total_factor(self):
    """we adjust the learning rate by size reduction."""
    ref_total = self.ref_a + self.ref_p
    trial_total = self.total_a_bits + self.total_p_bits
    return (trial_total - ref_total) / ref_total

  def print_stats(self):
    """Prints statistics of current model."""
    str_format = (
        "stats: delta_p={} delta_n={} rate={} trial_size={} reference_size={}\n"
        "       delta={:.2f}%"
    )

    print(
        str_format.format(
            self.delta_p, self.delta_n, self.rate, self.trial_size,
            int(self.reference_size), 100*self.delta())
    )

    a_percentage = np.round(
        100.0 * (self.total_a_bits - self.ref_a) / self.ref_a, 2)
    p_percentage = np.round(
        100.0 * (self.total_p_bits - self.ref_p) / self.ref_p, 2)
    ref_total = self.ref_a + self.ref_p
    trial_total = self.total_a_bits + self.total_p_bits
    total_percentage = np.round(
        100.0 * (trial_total - ref_total) / ref_total, 2)

    print(
        (
            "       a_bits={}/{} ({:.2f}%) p_bits={}/{} ({:.2f}%)\n"
            "       total={}/{} ({:.2f}%)"
        ).format(
            int(self.total_a_bits), int(self.ref_a), a_percentage,
            int(self.total_p_bits), int(self.ref_p), p_percentage,
            int(trial_total), int(ref_total), total_percentage
        ))



# Lint as: python3
# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Create multiplier quantizer."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import logging
from qkeras.qtools.quantized_operators import multiplier_impl
from qkeras.qtools.quantized_operators import quantizer_impl


class MultiplierFactory:
  """determine which multiplier implementation to use."""

  def __init__(self):
    # the table below is found in this slides:
    # https://docs.google.com/presentation/d/1pcmoB6ZpX0IqjhSwgzO-oQwpMRYwIcDe/edit#slide=id.p40
    # also attached the output datatype in the table
    self.multiplier_impl_table = [
        [
            (
                multiplier_impl.FixedPointMultiplier,
                quantizer_impl.QuantizedBits()
            ),
            (multiplier_impl.Shifter, quantizer_impl.QuantizedBits()),
            (multiplier_impl.Mux, quantizer_impl.QuantizedBits()),
            (multiplier_impl.Mux, quantizer_impl.QuantizedBits()),
            (multiplier_impl.AndGate, quantizer_impl.QuantizedBits()),
            (
                multiplier_impl.FloatingPointMultiplier,
                quantizer_impl.FloatingPoint(
                    bits=None)
            )
        ],
        [
            (multiplier_impl.Shifter, quantizer_impl.QuantizedBits()),
            (multiplier_impl.Adder, quantizer_impl.PowerOfTwo()),
            (multiplier_impl.Mux, quantizer_impl.PowerOfTwo()),
            (multiplier_impl.Mux, quantizer_impl.PowerOfTwo()),
            (multiplier_impl.AndGate, quantizer_impl.PowerOfTwo()),
            (multiplier_impl.FloatingPointMultiplier,
             quantizer_impl.FloatingPoint(bits=None)
            )
        ],
        [
            (multiplier_impl.Mux, quantizer_impl.QuantizedBits()),
            (multiplier_impl.Mux, quantizer_impl.PowerOfTwo()),
            (multiplier_impl.Mux, quantizer_impl.Ternary()),
            (multiplier_impl.Mux, quantizer_impl.Ternary()),
            (multiplier_impl.AndGate, quantizer_impl.Ternary()),
            (multiplier_impl.FloatingPointMultiplier,
             quantizer_impl.FloatingPoint(bits=None))
        ],
        [
            (multiplier_impl.Mux, quantizer_impl.QuantizedBits()),
            (multiplier_impl.Mux, quantizer_impl.PowerOfTwo()),
            (multiplier_impl.Mux, quantizer_impl.Ternary()),
            (multiplier_impl.XorGate, quantizer_impl.Binary(
                use_01=False)),
            (multiplier_impl.AndGate, quantizer_impl.Ternary()),
            (multiplier_impl.FloatingPointMultiplier,
             quantizer_impl.FloatingPoint(bits=None))
        ],
        [
            (multiplier_impl.AndGate, quantizer_impl.QuantizedBits()),
            (multiplier_impl.AndGate, quantizer_impl.PowerOfTwo()),
            (multiplier_impl.AndGate, quantizer_impl.Ternary()),
            (multiplier_impl.AndGate, quantizer_impl.Ternary()),
            (multiplier_impl.AndGate, quantizer_impl.Binary(
                use_01=True)),
            (multiplier_impl.FloatingPointMultiplier,
             quantizer_impl.FloatingPoint(bits=None))
        ],
        [
            (
                multiplier_impl.FloatingPointMultiplier,
                quantizer_impl.FloatingPoint(bits=None)
            ),
            (
                multiplier_impl.FloatingPointMultiplier,
                quantizer_impl.FloatingPoint(bits=None)
            ),
            (
                multiplier_impl.FloatingPointMultiplier,
                quantizer_impl.FloatingPoint(bits=None)
            ),
            (
                multiplier_impl.FloatingPointMultiplier,
                quantizer_impl.FloatingPoint(bits=None)
            ),
            (
                multiplier_impl.FloatingPointMultiplier,
                quantizer_impl.FloatingPoint(bits=None)
            ),
            (
                multiplier_impl.FloatingPointMultiplier,
                quantizer_impl.FloatingPoint(bits=None)
            )
        ]
    ]

  def make_multiplier(
      self, weight_quantizer: quantizer_impl.IQuantizer,
      input_quantizer: quantizer_impl.IQuantizer
  ) -> multiplier_impl.IMultiplier:
    """Create a multiplier instance.

    The type and bit width of the multiplier is deteremined from the
    quantizer type of both the kernel (weight) and input tensor.

    The table below illustrates the rule of inferring multiplier type from the
    quantizer type of both the kernel (weight) and input tensor

                                        x
                      qb(n)   +/-,exp  t(-1,0,+1) b(-1,+1) b(0,1) float32
        qb(n)            *     << >>,-     ?,-       ?,-       ?
        +/-,exp        << >>,-   +         ?,-        ^      ?,-
      w t(-1,0,+1)      ?,-     ?,-        ?,^       ?,^      ^
        b(-1,+1)        ?,-      ^         ?,^        ^       ^
        b(0,1)           ?      ?,-         ^         ^       ^      &
        float32

    Args:
      weight_quantizer: weight quantizer type
      input_quantizer: input quantizer type

    Returns:
      An IMultiplier instance.
    """

    assert weight_quantizer is not None
    assert input_quantizer is not None

    (multiplier_impl_class, output_quantizer) = self.multiplier_impl_table[
        weight_quantizer.mode][input_quantizer.mode]

    logging.debug(
        "multiplier implemented as class %s",
        multiplier_impl_class.implemented_as())

    assert issubclass(multiplier_impl_class, multiplier_impl.IMultiplier)

    return multiplier_impl_class(
        weight_quantizer,
        input_quantizer,
        output_quantizer
    )



# Lint as: python3
# ==============================================================================
# Copyright 2020 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Implements forgiving factor metrics for energy consumption."""

import json
import numpy as np
from qkeras.autoqkeras.forgiving_metrics.forgiving_factor import ForgivingFactor   # pylint: disable=line-too-long
from qkeras.qtools import run_qtools
from qkeras.qtools import settings as qtools_settings


class ForgivingFactorPower(ForgivingFactor):
  """Get Power cost of a given model."""

  def __init__(self, delta_p, delta_n, rate, stress=1.0, **kwargs):

    # input parameters:
    # delta_p, delta_n, rate: same as parent class
    # stress: stress level to shift reference curve
    # process: technology process to use in configuration (horowitz, ...)
    # parameters_on_memory: whether to store parameters in dram, sram, or fixed
    # activations_on_memory: store activations in dram, sram
    # min_sram_size: minimum sram size in number of bits
    # rd_wr_on_io: whether load data from dram to sram (consider sram as a cache
    #   for dram. If false, we will assume data will be already in SRAM
    # config_json: if None, use qtools/config_json by default
    #   define default source quantizers;
    #   default quantizers for intermediate variables if no quantizer provided
    #   parameters for energy calculation
    # source_quantizers: quantizer for model input
    # trained_model: whether model has been trained already, which is
    #   needed to compute tighter bounds for qBatchNorm Power estimation.
    # reference_internal: size to use for weight/bias/activation in
    #   get_reference energy calculation (int8, fp16, fp32)
    # reference_accumulator: accumulator and multiplier type in get_reference
    #   energy calculation
    # keras_layer_quantizer: quantizer for keras layers in hybrid models

    super(ForgivingFactorPower, self).__init__(delta_p, delta_n, rate)

    self.stress = stress
    # process: horowitz... - must be present in config_json
    self.process = kwargs.get("process", "horowitz")
    # parameters_on_memory: fixed, sram, dram
    self.parameters_on_memory = kwargs.get(
        "parameters_on_memory", ["fixed"] * 2)
    # activations_on_memory: sram, dram
    self.activations_on_memory = kwargs.get(
        "activations_on_memory", ["dram"] * 2
    )
    self.min_sram_size = kwargs.get("min_sram_size", [0] * 2)
    # rd_wr_on_io: true/false
    self.rd_wr_on_io = kwargs.get("rd_wr_on_io", [True] * 2)
    self.config_json = kwargs.get("config_json", None)
    self.source_quantizers = kwargs.get("source_quantizers", None)
    # trained_model: true/false
    self.trained_model = kwargs.get("trained_model", False)
    # reference_internal: int8, fp16, fp32
    self.reference_internal = kwargs.get("reference_internal", "fp32")
    # reference_internal: int8, int16, int32, fp16, fp32
    self.reference_accumulator = kwargs.get("reference_accumulator", "fp32")

    self.reference_size = None

    # energy_dict is a dictionary that lists energy consumption for each layer
    # format:
    #  {
    #     "layer0_name":
    #     {
    #        "mem_cost": 148171,
    #        "op_cost": 0
    #     },
    #     "layer1_name":
    #     {
    #         "mem_cost": 179923,
    #         "op_cost": 34
    #     },
    #     ...
    #
    #     "total_cost": 328129
    #  }

    self.ref_energy_dict = None
    self.trial_energy_dict = None

    assert self.parameters_on_memory[0] in ["dram", "sram", "fixed"]
    assert self.parameters_on_memory[1] in ["dram", "sram", "fixed"]
    assert self.activations_on_memory[0] in ["dram", "sram", "fixed"]
    assert self.activations_on_memory[1] in ["dram", "sram", "fixed"]
    assert self.reference_internal in ["fp16", "fp32", "int8"]
    assert self.reference_accumulator in ["int16", "int32", "fp16", "fp32"]

  def get_reference(self, model):
    # we only want to compute reference once
    if self.reference_size is not None:
      return self.reference_size * self.stress

    q = run_qtools.QTools(
        model, process=self.process,
        source_quantizers=self.reference_internal,
        is_inference=self.trained_model,
        weights_path=None,
        keras_quantizer=self.reference_internal,
        keras_accumulator=self.reference_accumulator,
        for_reference=True)

    energy_dict = q.pe(
        weights_on_memory=self.parameters_on_memory[0],
        activations_on_memory=self.activations_on_memory[0],
        min_sram_size=self.min_sram_size[0],
        rd_wr_on_io=self.rd_wr_on_io[0])

    self.ref_energy_dict = energy_dict
    self.reference_size = q.extract_energy_sum(
        qtools_settings.cfg.include_energy, energy_dict)

    self.reference_energy_profile = q.extract_energy_profile(
        qtools_settings.cfg.include_energy, energy_dict)

    return self.reference_size * self.stress

  def get_trial(self, model):
    """Computes size of quantization trial."""

    q = run_qtools.QTools(
        model, process=self.process,
        source_quantizers=self.source_quantizers,
        is_inference=self.trained_model,
        weights_path=None,
        keras_quantizer=self.reference_internal,
        keras_accumulator=self.reference_accumulator,
        for_reference=False)

    energy_dict = q.pe(
        weights_on_memory=self.parameters_on_memory[1],
        activations_on_memory=self.activations_on_memory[1],
        min_sram_size=self.min_sram_size[1],
        rd_wr_on_io=self.rd_wr_on_io[1])

    self.trial_energy_dict = energy_dict
    # self.trial_size = energy_dict["total_cost"]
    self.trial_size = q.extract_energy_sum(
        qtools_settings.cfg.include_energy, energy_dict)

    self.trial_energy_profile = q.extract_energy_profile(
        qtools_settings.cfg.include_energy, energy_dict)

    return self.trial_size

  def get_total_factor(self):
    """we adjust the learning rate by size reduction."""
    return (self.trial_size - self.reference_size) / self.reference_size

  def get_reference_stats(self):
    return self.reference_energy_profile

  def get_trial_stats(self):
    return self.trial_energy_profile

  def print_stats(self, verbosity=0):
    """Prints statistics of current model."""

    delta = self.delta()

    if (self.ref_energy_dict and self.trial_energy_dict):
      str_format = (
          "stats: delta_p={} delta_n={} rate={} trial_size={} "
          "reference_size={}\n"
          "       delta={:.2f}%"
      )

      print(
          str_format.format(
              self.delta_p, self.delta_n, self.rate, self.trial_size,
              int(self.reference_size), 100 * delta)
      )

    if verbosity > 0 and self.ref_energy_dict:
      print("Reference Cost Distribution:")
      dict_to_json = json.dumps(self.ref_energy_dict, indent=4)
      print(dict_to_json)

    if verbosity > 0 and self.trial_energy_dict:
      print("Trial Cost Distribution:")
      dict_to_json = json.dumps(self.trial_energy_dict, indent=4)
      print(dict_to_json)

    if (self.ref_energy_dict and self.trial_energy_dict):
      print("Total Cost Reduction:")
      reduction_percentage = np.round(
          100.0 * (self.trial_size - self.reference_size) /
          self.reference_size, 2)

      print(
          ("       {} vs {} ({:.2f}%)").format(
              int(self.trial_size), int(self.reference_size),
              reduction_percentage
          ))



# Copyright 2020 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utility functions for folding batchnorm with qconv/qdense layers."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import networkx as nx
import tensorflow as tf
from six.moves import range
from tensorflow.keras.models import clone_model
from tensorflow.keras.models import Model
from tensorflow.keras import Input

from .qconvolutional import QConv2D
from .qconvolutional import QDepthwiseConv2D
from .qtools import generate_layer_data_type_map as gen_map
from .qtools import qgraph


def convert_folded_layer_to_unfolded(layer):
  """Replace a source batchnorm folded layer with a non-folded layer.

  Args:
    layer: keras/qkeras layer type. Source layer to be replaced with

  Returns:
    new layer instance
  """

  # get layer config from the composite layer
  config = layer.get_config()
  # set layer config for QConv2D layer by first creating a tmp
  # QConv2D object and generate template for its config
  if layer.__class__.__name__ == "QConv2DBatchnorm":
    new_layer = QConv2D(filters=1, kernel_size=(2, 2), use_bias=True)
  elif layer.__class__.__name__ == "QDepthwiseConv2DBatchnorm":
    new_layer = QDepthwiseConv2D(kernel_size=(2, 2), use_bias=True)
  else:
    # TODO(lishanok): will extend to QDense in the future
    assert ValueError, "%s is not supported!" % layer.__class__.__name__

  new_layer_cfg = new_layer.get_config()

  # set qconv2d config according to the values in the composite layer
  for (key, _) in new_layer_cfg.items():
    if key in config.keys():
      new_layer_cfg[key] = config[key]

  # in case use_bias is False in the composite layer,
  #  we need to set it True because we have folded bias
  new_layer_cfg["use_bias"] = True

  # create a non-folded, e.g., qconv2d layer from config and replace
  # old layer with it
  if layer.__class__.__name__ == "QConv2DBatchnorm":
    new_layer = QConv2D.from_config(new_layer_cfg)
  elif layer.__class__.__name__ == "QDepthwiseConv2DBatchnorm":
    new_layer = QDepthwiseConv2D.from_config(new_layer_cfg)
  else:
    raise ValueError("Unsupported layer conversion {}".format(layer.name))

  return new_layer


def unfold_model(model):
  """Convert a model with batchnorm folded layer to a normal model.

  "Normal" here refers to a model without composite folded layer such as
  QConv2DBatchnorm layer.
  This function replace the folded layers with a normal QConv/QDense
  layer. It aslo sets the weights in the normal layer with the folded weights
  in the folded layer. Model architecture could be either sequential or
  non-sequential.

  Arguments:
    model: keras object, model with folded layers.

  Returns:
    A model that replaces folded layers (e.g., QConv2DBatchnorm) with normal
      qkeras layers (e.g., QConv2D). This model can be passed on to hardware
      generator so that hardware doesn't see batch normalization
      parameters.
  """

  def _convert_folded_layer(layer):
    if layer.__class__.__name__ in [
        "QConv2DBatchnorm", "QDepthwiseConv2DBatchnorm"]:
      new_layer = convert_folded_layer_to_unfolded(layer)
    else:
      new_layer = layer.__class__.from_config(layer.get_config())

    new_layer.build(layer.input_shape)
    return new_layer

  def _clone_weights(src_layer, new_layer):
    if (src_layer.__class__.__name__ == "QConv2DBatchnorm") and (
        new_layer.__class__.__name__ == "QConv2D"):
      src_weights = src_layer.get_folded_weights()
      # transfer weights from folded layer to the target layer
      folded_kernel_quantized = (
          src_weights[0].numpy())
      folded_bias_quantized = (
          src_weights[1].numpy())
      new_layer.set_weights([folded_kernel_quantized, folded_bias_quantized])

    elif (src_layer.__class__.__name__ == "QDepthwiseConv2DBatchnorm") and (
        new_layer.__class__.__name__ == "QDepthwiseConv2D"):
      # transfer weights from folded layer to the target layer
      src_weights = src_layer.get_folded_weights()
      folded_depthwise_kernel_quantized = src_weights[0].numpy()
      folded_bias_quantized = src_weights[1].numpy()
      new_layer.set_weights(
          [folded_depthwise_kernel_quantized, folded_bias_quantized])
    else:
      new_layer.set_weights(src_layer.get_weights())

  inp = Input(shape=model.input_shape[1:])
  cloned_model = clone_model(
      model, input_tensors=inp, clone_function=_convert_folded_layer)

  # replace weights
  for (src_layer, new_layer) in zip(model.layers, cloned_model.layers):
    _clone_weights(src_layer, new_layer)

  return cloned_model


def populate_bias_quantizer_from_accumulator(model, source_quantizers):
  """Populate the bias quantizer from accumulator type.

  When user set bias_quantizer=None for layers(e.g.,
  QConv2DBatchnorm), this function generates the accumulator type of
  the layer MAC op and set it as the bias quantizer.
  Such step is skipped if user provided a specific bias quantizer type.

  Args:
    model: keras/qkeras model object. If the model doesn't contain any batchnorm
      folded layer or if the bias quanizer type in the folded layer is already
      given, no operation needed. Else we generate the bias quantizer type and
      set it in model.

    source_quantizers: list of qkeras quantizers. A list of quantizer types
      for model inputs.

  Returns:
    keras model object
  """
  default_quantizer = "quantized_bits(8, 0, 1)"

  # if source_quantizers is None, CreateGraph will use default_quantizer
  (graph, source_quantizer_list) = qgraph.CreateGraph(
      model, source_quantizers, default_quantizer)
  qgraph.GraphPropagateActivationsToEdges(graph)

  # generate the quantizer types of each layer. For folded layers, if bias
  # quantizer is not given by user, this function will generate the accumulator
  # type and set it as the bias quantizer type.
  is_inference = False
  keras_quantizer = "quantized_bits(8, 0, 1)"
  keras_accumulator = "quantized_bits(8, 0, 1)"
  for_reference = False
  layer_map = gen_map.generate_layer_data_type_map(
      graph, source_quantizer_list, is_inference,
      keras_quantizer, keras_accumulator, for_reference)

  for layer in model.layers:
    # TODO(lishanok): extend to other layer types if necessary
    if layer.__class__.__name__ in [
        "QConv2DBatchnorm", "QDepthwiseConv2DBatchnorm"]:
      if not layer.bias_quantizer:
        # if user didn't specify the bias quantizer, we set it as the
        # MAC accumulator type of the current layer's MAC operation
        qtools_bias_quantizer = layer_map["layer_data_type_map"][
            layer].bias_quantizer

        if tf.is_tensor(qtools_bias_quantizer.int_bits):
          qtools_bias_quantizer.int_bits = (
              qtools_bias_quantizer.int_bits.numpy())

        layer.bias_quantizer = (
            qtools_bias_quantizer.convert_to_qkeras_quantizer())

        layer.bias_quantizer_internal = layer.bias_quantizer
        if layer.__class__.__name__ == "QConv2DBatchnorm":
          layer.quantizers = [layer.kernel_quantizer_internal,
                              layer.bias_quantizer_internal]
        elif layer.__class__.__name__ == "QDepthwiseConv2DBatchnorm":
          layer.quantizers = [layer.depthwise_quantizer_internal,
                              layer.bias_quantizer_internal]
  return model



# Lint as: python3
# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""quantized batch normliaztion quantizer implementation."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
import math

import numpy as np
from qkeras.qtools.quantized_operators import adder_factory
from qkeras.qtools.quantized_operators import divider_factory
from qkeras.qtools.quantized_operators import multiplier_factory
from qkeras.qtools.quantized_operators import quantizer_impl


class QBNFactory:
  """determine which quantizer implementation to use.

  Create an qbn instance. The type and bit width of the output_quantizer
  is deteremined from gamma, beta, mean and variance quantizer
  y = gamma * (x - mean)/stddev + beta
  """

  def make_quantizer(
      self, input_quantizer: quantizer_impl.IQuantizer,
      gamma_quantizer: quantizer_impl.IQuantizer,
      beta_quantizer: quantizer_impl.IQuantizer,
      mean_quantizer: quantizer_impl.IQuantizer,
      variance_quantizer: quantizer_impl.IQuantizer,
      use_scale,
      use_center
  ):
    """make a qbn quantizer."""

    self.input_quantizer = input_quantizer
    self.gamma_quantizer = gamma_quantizer
    self.beta_quantizer = beta_quantizer
    self.mean_quantizer = mean_quantizer
    self.variance_quantizer = variance_quantizer
    self.use_scale = use_scale
    self.use_center = use_center

    multiplier = None
    accumulator = None

    # convert variance po2 quantizer to stddev po2 quantizer
    stddev_quantizer = copy.deepcopy(variance_quantizer)
    if stddev_quantizer.is_po2:
      if variance_quantizer.max_val_po2 >= 0:
        stddev_quantizer.max_val_po2 = np.round(math.sqrt(
            variance_quantizer.max_val_po2))
      else:
        stddev_quantizer.max_val_po2 = variance_quantizer.max_val_po2

      stddev_quantizer.bits = variance_quantizer.bits - 1
      stddev_quantizer.int_bits = stddev_quantizer.bits

    divider_instance = divider_factory.IDivider()

    if use_scale:
      # gamma/var
      divider = divider_instance.make_quantizer(
          gamma_quantizer, stddev_quantizer)

      # update the actual number of values in divider quantizer during inference
      count = -1
      if gamma_quantizer.is_po2 and gamma_quantizer.inference_value_counts > 0:
        count = gamma_quantizer.inference_value_counts
        if stddev_quantizer.is_po2 and stddev_quantizer.inference_value_counts > 0:
          count *= stddev_quantizer.inference_value_counts
        else:
          count = -1
      if count > 0:
        divider.output.inference_value_counts = count

      # gamma/var * x
      multiplier_instance = multiplier_factory.MultiplierFactory()
      multiplier = multiplier_instance.make_multiplier(
          divider.output, input_quantizer)
      accumulator_input = multiplier

    else:
      # x/var
      divider = divider_instance.make_quantizer(
          input_quantizer, stddev_quantizer)
      accumulator_input = divider

    if use_center:
      # y = gamma/var * x + beta
      accumulator_instance = adder_factory.IAdder()
      accumulator = accumulator_instance.make_quantizer(
          accumulator_input.output, beta_quantizer)
      output_q = accumulator
    else:
      output_q = accumulator_input

    self.internal_divide_quantizer = divider
    self.internal_multiplier = multiplier
    self.internal_accumulator = accumulator
    self.internal_output = output_q



# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ==============================================================================
"""Definition of normalization quantization package."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import numpy as np
import six
import warnings

import tensorflow.compat.v2 as tf

from tensorflow.keras import constraints
from tensorflow.keras import initializers
from tensorflow.keras import regularizers
from tensorflow.keras.layers import BatchNormalization
from tensorflow.python.framework import ops
from tensorflow.python.framework import smart_cond as tf_utils
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import nn
from .qlayers import Clip
from .qlayers import get_auto_range_constraint_initializer
from .qlayers import get_quantizer
from .quantizers import quantized_relu_po2
from .quantizers import quantized_po2
from .safe_eval import safe_eval
from tensorflow_model_optimization.python.core.sparsity.keras.prunable_layer import PrunableLayer


class QBatchNormalization(BatchNormalization, PrunableLayer):
  """Quantized Batch Normalization layer.
  For training, mean and variance are not quantized.
  For inference, the quantized moving mean and moving variance are used.

  output = (x - mean) / sqrt(var + epsilon) * quantized_gamma + quantized_beta

  """

  def __init__(
      self,
      axis=-1,
      momentum=0.99,
      epsilon=1e-3,
      center=True,
      scale=True,
      activation=None,
      beta_initializer='zeros',
      gamma_initializer='ones',
      moving_mean_initializer='zeros',
      moving_variance_initializer='ones',
      beta_regularizer=None,
      gamma_regularizer=None,
      beta_quantizer='quantized_po2(5)',
      gamma_quantizer='quantized_relu_po2(6, 2048)',
      mean_quantizer='quantized_po2(5)',
      variance_quantizer='quantized_relu_po2(6, quadratic_approximation=True)',
      gamma_constraint=None,
      beta_constraint=None,
      # use quantized_po2 and enforce quadratic approximation
      # to get an even exponent for sqrt
      beta_range=None,
      gamma_range=None,
      **kwargs):

    if gamma_range is not None:
      warnings.warn('gamma_range is deprecated in QBatchNormalization layer.')

    if beta_range is not None:
      warnings.warn('beta_range is deprecated in QBatchNormalization layer.')

    self.gamma_range = gamma_range
    self.beta_range = beta_range
    self.activation = activation


    self.beta_quantizer = beta_quantizer
    self.gamma_quantizer = gamma_quantizer
    self.mean_quantizer = mean_quantizer
    self.variance_quantizer = variance_quantizer

    self.beta_quantizer_internal = get_quantizer(self.beta_quantizer)
    self.gamma_quantizer_internal = get_quantizer(self.gamma_quantizer)
    self.mean_quantizer_internal = get_quantizer(self.mean_quantizer)
    self.variance_quantizer_internal = get_quantizer(self.variance_quantizer)

    if hasattr(self.gamma_quantizer_internal, '_set_trainable_parameter'):
      self.gamma_quantizer_internal._set_trainable_parameter()
    if hasattr(self.variance_quantizer_internal, '_set_trainable_parameter'):
      self.variance_quantizer_internal._set_trainable_parameter()

    self.quantizers = [
        self.gamma_quantizer_internal,
        self.beta_quantizer_internal,
        self.mean_quantizer_internal,
        self.variance_quantizer_internal
    ]

    if scale and self.gamma_quantizer:
      gamma_constraint, gamma_initializer = (
          get_auto_range_constraint_initializer(
              self.gamma_quantizer_internal,
              gamma_constraint,
              gamma_initializer)
      )

    if center and self.beta_quantizer:
      beta_constraint, beta_initializer = (
          get_auto_range_constraint_initializer(
              self.beta_quantizer_internal,
              beta_constraint,
              beta_initializer)
      )

    if kwargs.get('fused', None):
      warnings.warn('batch normalization fused is disabled '
                    'in qkeras qnormalization.py.')
      del kwargs['fused']

    if kwargs.get('renorm', None):
      warnings.warn('batch normalization renorm is disabled '
                    'in qkeras qnormalization.py.')
      del kwargs['renorm']

    if kwargs.get('virtual_batch_size', None):
      warnings.warn('batch normalization virtual_batch_size is disabled '
                    'in qkeras qnormalization.py.')
      del kwargs['virtual_batch_size']

    if kwargs.get('adjustment', None):
      warnings.warn('batch normalization adjustment is disabled '
                    'in qkeras qnormalization.py.')
      del kwargs['adjustment']

    super(QBatchNormalization, self).__init__(
        axis=axis,
        momentum=momentum,
        epsilon=epsilon,
        center=center,
        scale=scale,
        beta_initializer=beta_initializer,
        gamma_initializer=gamma_initializer,
        moving_mean_initializer=moving_mean_initializer,
        moving_variance_initializer=moving_variance_initializer,
        beta_regularizer=beta_regularizer,
        gamma_regularizer=gamma_regularizer,
        beta_constraint=beta_constraint,
        gamma_constraint=gamma_constraint,
        fused=False,
        renorm=False,
        virtual_batch_size=None,
        adjustment=None,
        **kwargs)

  def call(self, inputs, training=None):
    if self.scale and self.gamma_quantizer:
      quantized_gamma = self.gamma_quantizer_internal(self.gamma)
    else:
      quantized_gamma = self.gamma

    if self.center and self.beta_quantizer:
      quantized_beta = self.beta_quantizer_internal(self.beta)
    else:
      quantized_beta = self.beta

    if self.mean_quantizer:
      quantized_moving_mean = self.mean_quantizer_internal(self.moving_mean)
    else:
      quantized_moving_mean = self.moving_mean

    if self.variance_quantizer:
      quantized_moving_variance = self.variance_quantizer_internal(
          self.moving_variance)
    else:
      quantized_moving_variance = self.moving_variance

    training = self._get_training_value(training)

    # Compute the axes along which to reduce the mean / variance
    input_shape = inputs.shape
    ndims = len(input_shape)
    reduction_axes = [i for i in range(ndims) if i not in self.axis]

    # Broadcasting only necessary for single-axis batch norm where the axis is
    # not the last dimension
    broadcast_shape = [1] * ndims
    broadcast_shape[self.axis[0]] = input_shape.dims[self.axis[0]].value
    def _broadcast(v):
      if (v is not None and len(v.shape) != ndims and
          reduction_axes != list(range(ndims - 1))):
        return array_ops.reshape(v, broadcast_shape)
      return v

    scale, offset = _broadcast(quantized_gamma), _broadcast(quantized_beta)

    # Determine a boolean value for `training`: could be True, False, or None.
    training_value = tf_utils.smart_constant_value(training)
    if training_value == False:  # pylint: disable=singleton-comparison,g-explicit-bool-comparison
      quantized_mean, quantized_variance = (quantized_moving_mean,
                                            quantized_moving_variance)
    else:
      # Some of the computations here are not necessary when training==False
      # but not a constant. However, this makes the code simpler.
      keep_dims = len(self.axis) > 1
      mean, variance = self._moments(
          math_ops.cast(inputs, self._param_dtype),
          reduction_axes,
          keep_dims=keep_dims)

      moving_mean = self.moving_mean
      moving_variance = self.moving_variance

      mean = tf_utils.smart_cond(
          training, lambda: mean, lambda: ops.convert_to_tensor(moving_mean))
      variance = tf_utils.smart_cond(
          training,
          lambda: variance,
          lambda: ops.convert_to_tensor(moving_variance))

      new_mean, new_variance = mean, variance

      if self.mean_quantizer:
        quantized_mean = self.mean_quantizer_internal(mean)
      else:
        quantized_mean = mean

      if self.variance_quantizer:
        quantized_variance = self.variance_quantizer_internal(variance)
      else:
        quantized_variance = variance

      if self._support_zero_size_input():
        inputs_size = array_ops.size(inputs)
      else:
        inputs_size = None

      def _do_update(var, value):
        """Compute the updates for mean and variance."""
        return self._assign_moving_average(var, value, self.momentum,
                                           inputs_size)

      def mean_update():
        true_branch = lambda: _do_update(self.moving_mean, new_mean)
        false_branch = lambda: self.moving_mean
        return tf_utils.smart_cond(training, true_branch, false_branch)

      def variance_update():
        """Update the moving variance."""
        true_branch = lambda: _do_update(self.moving_variance, new_variance)
        false_branch = lambda: self.moving_variance
        return tf_utils.smart_cond(training, true_branch, false_branch)

      self.add_update(mean_update)
      self.add_update(variance_update)

    quantized_mean = math_ops.cast(quantized_mean, inputs.dtype)
    quantized_variance = math_ops.cast(quantized_variance, inputs.dtype)
    if offset is not None:
      offset = math_ops.cast(offset, inputs.dtype)
    if scale is not None:
      scale = math_ops.cast(scale, inputs.dtype)
    # TODO(reedwm): Maybe do math in float32 if given float16 inputs, if doing
    # math in float16 hurts validation accuracy of popular models like resnet.
    outputs = nn.batch_normalization(inputs,
                                     _broadcast(quantized_mean),
                                     _broadcast(quantized_variance),
                                     offset,
                                     scale,
                                     self.epsilon)
    # If some components of the shape got lost due to adjustments, fix that.
    outputs.set_shape(input_shape)

    return outputs

  def get_config(self):
    config = {
        'axis': self.axis,
        'momentum': self.momentum,
        'epsilon': self.epsilon,
        'center': self.center,
        'scale': self.scale,
        'beta_quantizer':
            constraints.serialize(self.beta_quantizer_internal),
        'gamma_quantizer':
            constraints.serialize(self.gamma_quantizer_internal),
        'mean_quantizer':
            constraints.serialize(self.mean_quantizer_internal),
        'variance_quantizer':
            constraints.serialize(self.variance_quantizer_internal),
        'beta_initializer': initializers.serialize(self.beta_initializer),
        'gamma_initializer': initializers.serialize(self.gamma_initializer),
        'moving_mean_initializer':
            initializers.serialize(self.moving_mean_initializer),
        'moving_variance_initializer':
            initializers.serialize(self.moving_variance_initializer),
        'beta_regularizer': regularizers.serialize(self.beta_regularizer),
        'gamma_regularizer': regularizers.serialize(self.gamma_regularizer),
        'beta_constraint': constraints.serialize(self.beta_constraint),
        'gamma_constraint': constraints.serialize(self.gamma_constraint),
        'beta_range': self.beta_range,
        'gamma_range': self.gamma_range,
    }
    base_config = super(QBatchNormalization, self).get_config()
    return dict(list(base_config.items()) + list(config.items()))

  def compute_output_shape(self, input_shape):
    return input_shape

  def get_quantizers(self):
    return self.quantizers

  def get_prunable_weights(self):
    return []




# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Definition of quantization package."""

# Some parts of the code were taken from
#
# https://github.com/BertMoons/QuantizedNeuralNetworks-Keras-Tensorflow
#
# and follows several papers.
#
#    https://arxiv.org/pdf/1609.07061.pdf
#

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import defaultdict

import numpy as np
import tensorflow.compat.v1 as tf
from absl import logging

from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import InputLayer
from tensorflow.keras.models import Model

from .qlayers import QActivation
from .qlayers import QAdaptiveActivation
from .qlayers import QDense
from .qconvolutional import QConv1D
from .qconvolutional import QConv2D
from .qconvolutional import QDepthwiseConv2D
from .qconvolutional import QSeparableConv2D
from .qpooling import QAveragePooling2D
from .quantizers import quantized_bits
from .quantizers import quantized_relu
from .quantizers import quantized_tanh
from .quantizers import quantized_ulaw
from .bn_folding_utils import unfold_model
from .utils import get_model_sparsity


def analyze_accumulator(in_model, x, verbose=False):
  """Analyzes the distribution of weights to specify size of accumulators.

     Computes the maximum number of bits for the accumulator assuming the
     inputs have a distribution given by the dictionary x.

     for each output channel i:
       max_positive_value[i] = sum(w[i]) + bias[i] for the positive weights
       max_negative_value[i] = sum(w[i]) + bias[i] for the negative weights

     max_value = max(
            max_positive_value[i] * positive(x) +
            max_negative_value[i] * negative(x),

         - (max_negative_value[i] * positive(x) +
            max_positive_value[i] * negative(x))
     )

     accumulator_size = ceil( log2( max_value ) )

     x right now is a dictionary of the form:

     { layer_name: (min_value, max_value) }

     in the future, we want to provide a sample and compute this automatically

  Arguments:
    in_model: keras model object, model to be evaluated
    x: dictionary of the form: { layer_name: (min_value, max_value) }
       input distribution
    verbose: boolean, if true, print statistics messages

  Returns:
    dictionary containing { layer_name: accumulator_size }
  """

  # this function converts a folded model to a "normal" model. It replace folded
  # layers (e.g., QConv2dBatchnorm) layer with qconv2d layer whenever possible.
  model = unfold_model(in_model)

  acc_sizes = {}

  for layer in model.layers:
    if (isinstance(layer, QDepthwiseConv2D) or
        isinstance(layer, QConv2D) or
        isinstance(layer, QConv1D) or
        isinstance(layer, QDense)):
      weights = layer.get_weights()
      k = weights[0]
      if layer.use_bias:
        b = weights[1]
      else:
        b = np.zeros((k.shape[-1],), dtype=np.float32)

      all_bits = []
      nbits = []
      for i in range(k.shape[1]):
        # compute sum of positive weights
        npp = np.sum(k[..., i] * (k[..., i] > 0)) + (b[i] > 0) * b[i]

        # compute sum of negative weights
        nnn = np.sum(k[..., i] * (k[..., i] < 0)) + (b[i] < 0) * b[i]

        # largest value is
        #   npp * largest positive - nnn * largest_negative or
        #   nnn * largest_positive - npp * largest_negative

        x_min = x[layer.name][0]
        x_max = x[layer.name][1]

        n1 = npp * (x_max > 0) * x_max + nnn * (x_min < 0) * x_min
        n0 = - (nnn * (x_max > 0) * x_max + npp * (x_min < 0) * x_min)

        if n1 > n0:
          nbits.append(n1)
        else:
          nbits.append(n0)

        all_bits.append((n1, n0))

      max_bits = int(np.ceil(np.log2(max(nbits))))
      acc_sizes[layer.name] = max_bits

      if verbose:
        print()
        print(layer.name, "- input range:", x[layer.name])
        print("  max value:", np.amax(k))
        print("  min value:", np.amin(k))
        print("  most positive sum:", np.amax(np.array(all_bits)[:, 0]))
        print("  most negative sum:", -np.amax(np.array(all_bits)[:, 1]))
        print("  number of bits:", max_bits)

  if verbose:
    print()

  return acc_sizes


def analyze_accumulator_from_sample(
    in_model, x_sample, mode="conservative", verbose=False):
  """Extracts range of inputs of quantized layers from samples."""

  # mode is one of "conservative", "sampled"
  if mode not in ["conservative", "sampled"]:
    ValueError("'mode' has to be 'conservative' or 'sampled'")

  # this function converts a folded model to a "normal" model. It replace folded
  # layers (e.g., QConv2DBatchnorm) layer with qconv2d layer whenever possible.
  model = unfold_model(in_model)

  # get layer names of quantized layers (QDense and QConv2D)
  layer_names = [
      layer.name for layer in model.layers
      if (isinstance(layer, QDepthwiseConv2D) or isinstance(layer, QConv2D) or
          isinstance(layer, QConv1D) or isinstance(layer, QDense))
  ]

  # sampled mode: just apply x_sample and check the outputs

  if mode == "sampled":

    outputs = [
        layer.output for layer in model.layers
        if (isinstance(layer, QDepthwiseConv2D) or isinstance(layer, QConv2D) or
            isinstance(layer, QConv1D) or isinstance(layer, QDense))
    ]

    eval_outputs = Model(inputs=model.inputs, outputs=outputs)

    # predict values for all inputs to quantized layers

    values = eval_outputs.predict(x_sample)

    acc_sizes = {}

    for name, value in zip(layer_names, values):
      max_value = np.amax(np.abs(value))
      if max_value != 0:
        acc_sizes[name] = int(np.ceil(np.log2(max_value)))
      else:
        acc_sizes[name] = 0

    return acc_sizes

  # get inputs of quantized layers (QDense and QConv2D
  # we use Activation("linear") to trick keras and tensorflow
  # to avoid direct connections of inputs and any other
  # artifacts.

  outputs = [
      Activation("linear")(layer.input) for layer in model.layers
      if (isinstance(layer, QDepthwiseConv2D) or isinstance(layer, QConv2D) or
          isinstance(layer, QConv1D) or isinstance(layer, QDense))
  ]

  eval_inputs = Model(inputs=model.inputs, outputs=outputs)

  # predict values for all inputs to quantized layers

  values = eval_inputs.predict(x_sample)

  x_dict = {}

  for name, value in zip(layer_names, values):
    x_dict[name] = (np.amin(value), np.amax(value))

  return analyze_accumulator(model, x_dict, verbose)


def get_quant_mode(quant):
  """Returns the quantizer mode, number of bits and if it is a signed number."""

  #  qb(n)[0] +/-,exp[1] t(-1,0,+1)[2] b(-1,+1)[3] b(0,1)[4]
  #  entry is tuple:
  #    (instance name, mode #(above), number of bits (-1 means check class),
  #     sign bit)

  modes = [
      # depending on the number of bits, quantized_bits may be 2, 2
      ("quantized_bits", 0, -1, 1),
      ("bernoulli", 4, 1, 0),
      ("stochastic_ternary", 2, 2, 1),
      ("ternary", 2, 2, 1),
      ("stochastic_binary", 3, 1, 1),
      ("binary", 3, 1, 1),
      # depending on the number of bits, quantized_relu may be 4, 1
      ("quantized_relu", 0, -1, 0),
      # depending on the number of bits, quantized_tanh may be 2, 2
      ("quantized_ulaw", 0, -1, 1),
      ("quantized_tanh", 0, -1, 1),
      ("quantized_po2", 1, -1, 1),
      ("quantized_relu_po2", 1, -1, 0),
      ("float", 5, 32, 1)
  ]

  for (inst, mode, bits, sign) in modes:
    if not quant or getattr(quant, "__name__", None) == "linear":
      # if quantizer not specified or linear, we use float type
      if inst == "float":
        return (mode, bits, sign)

    elif quant.__class__.__name__ == inst:
      if bits == -1:
        bits = int(quant.bits)
        if (
            isinstance(quant, quantized_bits) or
            isinstance(quant, quantized_tanh) or
            isinstance(quant, quantized_ulaw)):
          if bits == 2 and int(quant.integer) == 1:
            mode = 2
        elif isinstance(quant, quantized_relu):
          if bits == 1 and int(quant.integer) == 1:
            mode = 4
      return (mode, bits, sign)
  raise ValueError("Quantizer {} Not Found".format(quant))


def get_operation_type(layer, output_cache):
  """Checks quantizers around layer and weights to get operation type.

  Determines operator strenght according to the following table.
                                      x
                     qb(n)   +/-,exp  t(-1,0,+1) b(-1,+1) b(0,1) float
      qb(n)            *     << >>,-     ?,-       ?,-       ?    *
      +/-,exp        << >>,-   +         ?,-        ^      ?,-    *
    w t(-1,0,+1)      ?,-     ?,-        ?,^       ?,^      ^     *
      b(-1,+1)        ?,-      ^         ?,^        ^       ^     *
      b(0,1)           ?      ?,-         ^         ^       ^     *
      float           *        *          *        *        *     *

  Arguments:
    layer: layer in Keras to determine the operation strength.
    output_cache: cache of input tensor bit sizes.

  Returns:
    One of "mult", "fmult", "adder", "barrel", "mux", "xor".
    Note: "mult" represents quantized bit multiplier, "fmult" represents
          floating point multiplier.
  """

  wx_table = [
      ["mult", "barrel", "mux", "mux", "mux", "fmult"],
      ["barrel", "adder", "mux", "xor", "mux", "fmult"],
      ["mux", "mux", "mux", "mux", "xor", "fmult"],
      ["mux", "xor", "mux", "xor", "xor", "fmult"],
      ["mux", "mux", "xor", "xor", "xor", "fmult"],
      ["fmult", "fmult", "fmult", "fmult", "fmult", "fmult"],
  ]

  # check if this is a quantized layers (QDense, QConv, QDepthwise)
  if hasattr(layer, "get_quantizers"):
    w_quant = layer.get_quantizers()[0]
    w_mode, w_bits, w_sign = get_quant_mode(w_quant)
    if w_mode == "float":
      logging.warning("%s kernel is unquantized!", layer.name)

    # for the input, get tensor input and search the cache that associates
    # the quantizer with a tensor
    if output_cache.get(layer.input.experimental_ref(), None) is not None:
      x_mode, x_bits, x_sign = get_quant_mode(
          output_cache.get(layer.input.experimental_ref()))
      if x_mode == "float":
        logging.warning("%s input is unquantized!", layer.name)
    else:
      print("cannot determine presently model for {}".format(layer.name))
      return "null", (w_mode, -1), (w_bits, -1), (w_sign, -1)
    mode = wx_table[w_mode][x_mode]
    return mode, (w_mode, x_mode), (w_bits, x_bits), (w_sign, x_sign)

  raise ValueError("Cannot find suitable quantization candidates for {}".format(
      layer.name))


def create_activation_cache(model):
  """Creates an activation cache for the tensors of a model."""

  input_quantizer = quantized_relu(8, 0)

  output_cache = {}

  # If using a Sequential model, the input layer is hidden. Therefore, add the
  # input quantization to the cache if the first layer is not an input layer
  if not isinstance(model.layers[0], InputLayer):
    output_cache[model.layers[0].input.experimental_ref()] = input_quantizer

  # cache graph tensors' activations

  for l in model.layers:
    output_cache[l.output.experimental_ref()] = l
    if isinstance(l, QActivation) or isinstance(l, QAdaptiveActivation) :
      output_cache[l.output.experimental_ref()] = l.quantizer
    elif isinstance(l, InputLayer):
      # assume the input is 8-bit positive value
      output_cache[l.output.experimental_ref()] = input_quantizer
    elif l.__class__.__name__ in [
        "QDense", "QConv2D", "QConv1D", "QDepthwiseConv2D"
    ]:
      output_cache[l.output.experimental_ref()] = l.activation
    else:
      if isinstance(l.input, list):
        # right now, we just get the first one - we assume this is the leading
        # one.
        all_q = [
            output_cache.get(l.input[i].experimental_ref())
            for i in range(len(l.input))
        ]
        q = all_q[0]
      else:
        q = output_cache.get(l.input.experimental_ref(), None)
      output_cache[l.output.experimental_ref()] = q
      if q is None:
        raise ValueError("Unknown operation in {}".format(l.name))

  return output_cache


def extract_model_operations(in_model):
  """Determines types of operations for convolutions."""

  model = unfold_model(in_model)
  cache_q = create_activation_cache(model)
  cache_o = {}

  operations = {}

  for layer in model.layers:

    if layer.__class__.__name__ == "InputLayer":
      continue

    if isinstance(layer.input, list):
      input_shape = [
          cache_o.get(layer.input[i].experimental_ref(),
                      layer.input[i].get_shape())
          for i in range(len(layer.input))
      ]
    else:
      input_shape = cache_o.get(layer.input.experimental_ref(),
                                layer.input.get_shape())

    # Check if the inputs are a list of Dimensions
    if isinstance(input_shape, list):
      # Iterate though all of the input shapes and extract the dimension values
      for i, dim in enumerate(input_shape):
        if isinstance(dim[0], tf.Dimension):
          shape = [None]
          for j in range(1, len(dim)):
            shape.append(dim[j] if isinstance(dim[j], int) else dim[j].value)
          input_shape[i] = tuple(shape)

    output_shape = layer.compute_output_shape(input_shape)

    cache_o[layer.output.experimental_ref()] = output_shape

    if layer.__class__.__name__ not in ["QDense", "QConv2D", "QConv1D",
                                        "QDepthwiseConv2D", "QSeparableConv1D",
                                        "QSeparableConv2D"]:
      continue

    if layer.__class__.__name__ in ["QConv2D"]:

      _, _, _, channels_i = input_shape

      _, height_o, width_o, channels_o = output_shape

      weight = layer.get_weights()[0]


      kernel_h, kernel_w, _, _ = weight.shape

      number_of_operations = (
          height_o * width_o * channels_o * kernel_h * kernel_w * channels_i)

      number_of_weights = (kernel_h * kernel_w * channels_o * channels_i)

      number_of_bias = 0
      if len(layer.get_weights()) > 1:
        number_of_bias = layer.get_weights()[1].shape[0]

      weight_quant, bias_quant = layer.get_quantizers()
      weight_type = get_quant_mode(weight_quant)
      bias_type = get_quant_mode(bias_quant)

      if weight_type[0] == "float":
        logging.warning("%s kernel is unquantized!", layer.name)
      if bias_type[0] == "float":
        logging.warning("%s bias is unquantized!", layer.name)

    elif layer.__class__.__name__ in ["QConv1D"]:

      _, _, channels_i = input_shape

      _, time_o, channels_o = output_shape

      weight = layer.get_weights()[0]

      kernel_length, _, _ = weight.shape

      number_of_operations = (
          time_o * channels_o * kernel_length * channels_i)

      number_of_weights = (kernel_length * channels_o * channels_i)
      number_of_bias = 0
      if len(layer.get_weights()) > 1:
        number_of_bias = layer.get_weights()[1].shape[0]

      weight_quant, bias_quant = layer.get_quantizers()
      weight_type = get_quant_mode(weight_quant)
      bias_type = get_quant_mode(bias_quant)

      if weight_type[0] == "float":
        logging.warning("%s kernel is unquantized!", layer.name)
      if bias_type[0] == "float":
        logging.warning("%s bias is unquantized!", layer.name)

    elif layer.__class__.__name__ in ["QDepthwiseConv2D"]:

      _, _, _, channels_i = input_shape

      _, height_o, width_o, channels_o = output_shape

      weight_1 = layer.get_weights()[0]

      kernel_h, kernel_w, _, _ = weight_1.shape

      number_of_operations = (
          kernel_h * kernel_w * height_o * width_o * channels_i)

      number_of_weights = (kernel_h * kernel_w * channels_o * channels_i)

      number_of_bias = 0
      if len(layer.get_weights()) > 1:
        number_of_bias = layer.get_weights()[1].shape[0]

      weight_quant, bias_quant = layer.get_quantizers()
      weight_type = get_quant_mode(weight_quant)
      bias_type = get_quant_mode(bias_quant)

      if weight_type[0]  == "float":
        logging.warning("%s kernel is unquantized!", layer.name)
      if bias_type[0]  == "float":
        logging.warning("%s bias is unquantized!", layer.name)

    elif layer.__class__.__name__ in ["QSeparableConv1D"]:

      _, _, channels_i = input_shape

      _, time_o, channels_o = output_shape

      weight_1 = layer.get_weights()[0]

      kernel_length, _, _ = weight_1.shape

      number_of_operations = (
          kernel_length * time_o * channels_i + 
          time_o * channels_o)

      number_of_weights = [
        kernel_length * channels_i, 
        channels_o * channels_i]

      number_of_bias = 0
      if len(layer.get_weights()) > 2:
        number_of_bias = layer.get_weights()[2].shape[0]

      depthwise_quant, pointwise_quant, bias_quant = layer.get_quantizers()
      depthwise_type = get_quant_mode(depthwise_quant)
      pointwise_type = get_quant_mode(pointwise_quant)
      weight_type = [depthwise_type, pointwise_type]
      bias_type = get_quant_mode(bias_quant)

      if depthwise_type[0] == "float":
        logging.warning("%s depthwise kernel is unquantized!", layer.name)
      if pointwise_type[0] == "float":
        logging.warning("%s pointwise kernel is unquantized!", layer.name)
      if bias_type[0] == "float":
        logging.warning("%s bias is unquantized!", layer.name)

    elif layer.__class__.__name__ in ["QSeparableConv2D"]:

      _, _, _, channels_i = input_shape

      _, height_o, width_o, channels_o = output_shape

      weight_1 = layer.get_weights()[0]

      kernel_h, kernel_w, _, _ = weight_1.shape

      number_of_operations = (
          kernel_h * kernel_w * height_o * width_o * channels_i + 
          height_o * width_o * channels_o)

      number_of_weights = [
        kernel_h * kernel_w * channels_i,
        channels_o * channels_i]

      number_of_bias = 0
      if len(layer.get_weights()) > 2:
        number_of_bias = layer.get_weights()[2].shape[0]

      depthwise_quant, pointwise_quant, bias_quant = layer.get_quantizers()
      depthwise_type = get_quant_mode(depthwise_quant)
      pointwise_type = get_quant_mode(pointwise_quant)
      weight_type = [depthwise_type, pointwise_type]
      bias_type = get_quant_mode(bias_quant)

      if depthwise_type[0] == "float":
        logging.warning("%s depthwise kernel is unquantized!", layer.name)
      if pointwise_type[0] == "float":
        logging.warning("%s pointwise kernel is unquantized!", layer.name)
      if bias_type[0] == "float":
        logging.warning("%s bias is unquantized!", layer.name)

    elif layer.__class__.__name__ in ["QDense"]:

      _, size_i = input_shape
      _, size_o = output_shape

      number_of_operations = (size_i * size_o)

      number_of_weights = size_i * size_o
      number_of_bias = 0
      if len(layer.get_weights()) > 1:
        number_of_bias = layer.get_weights()[1].shape[0]

      weight_quant, bias_quant = layer.get_quantizers()
      weight_type = get_quant_mode(weight_quant)
      bias_type = get_quant_mode(bias_quant)

      if weight_type[0] == "float":
        logging.warnings("%s kernel is unquantized!", layer.name)
      if bias_type[0] == "float":
        logging.warnings("%s bias is unquantized!", layer.name)

    # "number_of_operations" is tensor_shape.Dimension type
    operations[layer.name] = {
        "type":
            get_operation_type(layer, cache_q),
        "number_of_operations":
            number_of_operations if isinstance(number_of_operations, int) else
            number_of_operations.value,
        "number_of_weights":
            number_of_weights,
            # if isinstance(number_of_weights, int) else number_of_weights.value,
        "number_of_bias":
            number_of_bias,
            # if isinstance(number_of_bias, int) else number_of_bias.value,
        "type_of_weights":
            weight_type,
        "type_of_bias":
            bias_type,
    }

  return operations


def print_qstats(model):
  """Prints quantization statistics for the model."""

  model_ops = extract_model_operations(model)

  ops_table = defaultdict(lambda: 0)

  print("")
  print("Number of operations in model:")
  for name in sorted(model_ops):
    mode, _, sizes, signs = model_ops[name]["type"]
    number = model_ops[name]["number_of_operations"]
    sign = "s" if sum(signs) > 0 else "u"
    op_name = sign + mode + "_" + str(sizes[0]) + "_" + str(sizes[1])
    ops_table[op_name] += number
    print("    {:30}: {:5} ({})".format(str(name), str(number), str(op_name)))

  print("")
  print("Number of operation types in model:")
  for key in sorted(ops_table.keys()):
    if ops_table[key] > 0:
      print("    {:30}: {}".format(key, ops_table[key]))

  print("")
  print("Weight profiling:")
  for name in sorted(model_ops):
    weight_type = model_ops[name]["type_of_weights"]
    n_weights = model_ops[name]["number_of_weights"]
    if isinstance(weight_type, list):
      for i, (w_type, w_number) in enumerate(zip(weight_type, n_weights)):
        _, w_sizes, _ = w_type
        print("    {:30} : {:5} ({}-bit unit)".format(
            str(name) + "_weights_" + str(i), str(w_number), str(w_sizes)))
    else:
      _, w_sizes, _ = weight_type
      print("    {:30} : {:5} ({}-bit unit)".format(
          str(name) + "_weights", str(n_weights), str(w_sizes)))
    _, b_sizes, _ = model_ops[name]["type_of_bias"]
    b_number = model_ops[name]["number_of_bias"]
    print("    {:30} : {:5} ({}-bit unit)".format(
        str(name) + "_bias", str(b_number), str(b_sizes)))

  print("")
  print("Weight sparsity:")
  total_sparsity, per_layer = get_model_sparsity(model, per_layer=True)
  for layer in per_layer:
    print("    {:30} : {:.4f}".format(str(layer[0]), layer[1]))
  print("    " + ("-"*40))
  print("    {:30} : {:.4f}".format("Total Sparsity", total_sparsity))



# Copyright 2020 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

import os
import tensorflow as tf


class QNoiseScheduler(tf.keras.callbacks.Callback):
  """Schedules the gradual quantization noise training for each step (or epoch).

     It updates the qnoise_factor in the quantizers to gradually introduce the
     quantization noise during training.

     The idea was adopted from "https://arxiv.org/pdf/1903.01061.pdf"
  """

  def __init__(self,
               start,
               finish,
               freq_type="epoch",
               update_freq=1,
               initial_step_or_epoch=0,
               exponent=3.0,
               use_ste=True,
               log_dir=None):
    """Initializes this QNoiseScheduler.

    Args:
      start: Int. The step (epoch) to start the gradual training.
      finish: Int. The step (epoch) to finish the gradual training. When the
        start and the finish are equal, the qnoise_factor will be 1.0 in the
        beginning of the training.
      freq_type: Str. "step" or "epoch". It sets the qnoise_factor update
        frequency type.
      update_freq: Int. Updating frequency of the qnoise_factor.
      initial_step_or_epoch:  Int. Step or epoch at which to start training.
      exponent: Float. It is the exponent in the qnoise_factor calculation. It
        controls the rate of the gradual qnoise_factor change.
      use_ste: Bool. Whether to use "straight-through estimator" (STE) method or
        not.
      log_dir: Str. log directory to save qnoise_factor every epoch end.
    """
    super(QNoiseScheduler, self).__init__()

    self.start = start
    self.finish = finish
    if start > finish:
      raise ValueError(
          ("start {} must be greater than finish {}").format(start, finish))
    supported_freq_type = ["step", "epoch"]
    if freq_type not in supported_freq_type:
      raise ValueError(("Invalid frequency type {}. only {} are "
                        "supported.").format(freq_type, supported_freq_type))
    self.freq_type = freq_type
    self.update_freq = update_freq
    self.initial_step_or_epoch = initial_step_or_epoch
    self.exponent = exponent
    self.qnoise_factor = None
    self.use_ste = use_ste
    self.quantizers = None
    self.summary_writer = None
    if log_dir:
      self.summary_writer = tf.summary.create_file_writer(log_dir)
    self.num_iters = np.array(0, dtype="int64")

  def calculate_qnoise_factor(self, freq):
    """Returns calculated qnoise_factor based on the current step (epoch) and
    the schedule parameters.

    Args:
      freq: The current step (or epoch) to calculate the qnoise_factor.

    Returns:
      qnoise_factor : calculated qnoise_factor.
    """
    if freq < self.start:
      qnoise_factor = 0.0
    elif freq <= self.finish and self.start != self.finish:
      val = float(self.finish - freq) / float(self.finish - self.start)
      qnoise_factor = 1.0 - np.power(val, self.exponent)
    else:
      qnoise_factor = 1.0

    return qnoise_factor

  def set_qnoise_factor(self, quantizer, qnoise_factor):
    """Set self.qnoise_factor and update the qnoise_factor of the quantizer."""

    # Updating the qnoise_factor of the quantizer.
    quantizer.update_qnoise_factor(qnoise_factor)
    # Updating the qnoise_factor of the callback.
    self.qnoise_factor = qnoise_factor

  def set_quantizers(self):
    """Set quantizers to update the qnoise_factor.

    This must be called before building the quantizers.
    """
    for quantizer in self.quantizers:
      if hasattr(quantizer, "use_ste"):
        quantizer.use_ste = self.use_ste
      if hasattr(quantizer, "use_variables"):
        quantizer.use_variables = True
      if hasattr(quantizer, "built"):
        # If the quantizer has been built but not using tf.Variable then it
        # builds again to create tf.Variables.
        if quantizer.built and not isinstance(quantizer.qnoise_factor,
                                                 tf.Variable):
          quantizer.build(use_variables=True)

      # Set the qnoise_factor to 0.0 to pretrain without quantization.
      self.set_qnoise_factor(quantizer, qnoise_factor=0.0)

  def get_quantizers(self, model):
    """Returns a list of quantizers with qnoise_factor in the model.

    Args:
      model: model to get a list of quantizers with qnoise_factor.

    Returns:
      A list of quantizers with the qnoise_factor variable.
    """
    all_quantizers = []
    for layer in model.layers:
      # A list of attributes holding the quantizer(s).
      for attr in ["quantizers", "quantizer"]:
        if hasattr(layer, attr):
          quantizers = getattr(layer, attr)
          quantizers = quantizers if attr == "quantizers" else [quantizers]
          for quantizer in quantizers:
            if hasattr(quantizer, "qnoise_factor"):
              all_quantizers.append(quantizer)

    return all_quantizers

  def update_qnoise_factor(self, freq):
    """Update the qnoise_factor of the model.

    Args:
      freq: The current step (epoch) to calculate the qnoise_factor.
    """
    # Update the qnoise_factor at the frequency of self.update_freq.
    if freq % self.update_freq != 0:
      self.num_iters += 1
      return

    new_qnoise_factor = self.calculate_qnoise_factor(freq)
    for quantizer in self.quantizers:
      # Updates the qnoise factors of the quantizers in the model.
      self.set_qnoise_factor(quantizer, new_qnoise_factor)
    self.num_iters += 1

  def on_train_begin(self, logs=None):
    if not self.quantizers:
      # Build a list of quantizers which is used for updating qnoise_factor.
      self.quantizers = self.get_quantizers(self.model)
      self.set_quantizers()

  def on_epoch_begin(self, epoch, logs=None):
    if self.freq_type == "epoch":
      self.update_qnoise_factor(self.initial_step_or_epoch + self.num_iters)

  def on_epoch_end(self, epoch, logs=None):
    if self.summary_writer:
      with self.summary_writer.as_default():
        tf.summary.scalar("qnoise_factor", data=self.qnoise_factor, step=epoch)

  def on_train_batch_begin(self, batch, logs=None):
    if self.freq_type == "step":
      self.update_qnoise_factor(self.initial_step_or_epoch + self.num_iters)
